#!/usr/bin/env python3

# This file is part of Xpra.
# Copyright (C) 2023 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os.path
import shlex
import sys
import glob
import shutil

from xpra.os_util import WIN32
from xpra.util.io import get_status_output
from xpra.util.str_fn import sorted_nicely


def get_nvcc_version(command: str) -> tuple[int, ...]:
    if not os.path.exists(command):
        return None
    code, out, _ = get_status_output([command, "--version"])
    if code != 0:
        return None
    vpos = out.rfind(", V")
    if vpos > 0:
        version = out[vpos + 3:].split("\n")[0]
        version_str = f" version {version}"
    else:
        version = "0"
        version_str = " unknown version!"
    print(f"found CUDA compiler {command!r} : {version_str}")
    return tuple(int(x) for x in version.split("."))


def get_nvcc() -> [str, tuple[int, ...]]:
    path_options = os.environ.get("PATH", "").split(os.path.pathsep)
    if WIN32:
        nvcc_exe = "nvcc.exe"
        CUDA_DIR = os.environ.get("CUDA_DIR", "C:\\Program Files\\NVIDIA GPU Computing Toolkit\\CUDA")
        path_options += ["./cuda/bin/"] + list(reversed(sorted_nicely(glob.glob(f"{CUDA_DIR}\\*\\bin"))))
    else:
        nvcc_exe = "nvcc"
        path_options += ["/usr/local/cuda/bin", "/opt/cuda/bin"]
        path_options += list(reversed(sorted_nicely(glob.glob("/usr/local/cuda*/bin"))))
        path_options += list(reversed(sorted_nicely(glob.glob("/opt/cuda*/bin"))))
    options = [os.path.join(x, nvcc_exe) for x in path_options]
    # prefer the one we find on the $PATH, if any:
    v = shutil.which(nvcc_exe)
    if v and (v not in options):
        options.insert(0, v)
    nvcc_versions = {}
    for filename in options:
        vnum = get_nvcc_version(filename)
        if vnum:
            nvcc_versions[vnum] = filename
    if not nvcc_versions:
        print("unable to find nvcc")
        sys.exit(1)
    # choose the most recent one:
    nvcc_version, nvcc = list(reversed(sorted(nvcc_versions.items())))[0]
    if len(nvcc_versions) > 1:
        print(f" using version {nvcc_version} from {nvcc}")
    return nvcc, nvcc_version


def get_gcc_version() -> tuple[int, ...]:
    CC = os.environ.get("CC", "gcc")
    if CC.find("clang") >= 0:
        return (0,)
    exit_code, _, err = get_status_output([CC, "-v"])
    if exit_code != 0:
        return (0,)
    V_LINE = "gcc version "
    gcc_version = []
    for line in err.splitlines():
        if not line.startswith(V_LINE):
            continue
        v_str = line[len(V_LINE):].strip().split(" ")[0]
        for p in v_str.split("."):
            try:
                gcc_version.append(int(p))
            except ValueError:
                break
        break
    return tuple(gcc_version)


def get_nvcc_args(nvcc: str, nvcc_version=(0, 0)) -> list[str]:
    if nvcc_version < (11, 6):
        raise RuntimeError(f"nvcc version {nvcc_version} is too old, minimum is 11.6")
    nvcc_cmd = [
        nvcc,
        "-fatbin",
        "-std=c++14",
        "-arch=all",
        "-Wno-deprecated-gpu-targets",
        "-Xnvlink",
        "-ignore-host-info",
        "--allow-unsupported-compiler",
    ]
    if get_gcc_version() >= (14, 0):
        clangpp = shutil.which("clang++")
        if not clangpp:
            print("clang++ not found, compilation may fail")
        else:
            nvcc_cmd.append(f"-ccbin={clangpp}")
    return nvcc_cmd


def main(args) -> int:
    nvcc, nvcc_version = get_nvcc()
    nvcc_args = get_nvcc_args(nvcc, nvcc_version)
    if len(args) == 1:
        kernels = (
            "XRGB_to_NV12", "XRGB_to_YUV444", "BGRX_to_NV12", "BGRX_to_YUV444",
            "BGRX_to_RGB", "RGBX_to_RGB", "RGBA_to_RGBAP", "BGRA_to_RGBAP",
        )
    else:
        kernels = args[1:]
    nvcc_commands = []
    print("compiling CUDA kernels with:")
    print(" " + " ".join(f"{x!r}" for x in nvcc_args))
    for kernel in kernels:
        cuda_src = f"fs/share/xpra/cuda/{kernel}.cu"
        cuda_bin = f"fs/share/xpra/cuda/{kernel}.fatbin"
        kbuild_cmd = nvcc_args + ["-c", cuda_src, "-o", cuda_bin]
        print(f"* {kernel}")
        nvcc_commands.append(kbuild_cmd)
    # parallel build:
    nvcc_errors = []

    def nvcc_compile(nvcc_cmd: list[str]) -> None:
        c, stdout, stderr = get_status_output(nvcc_cmd)
        if c != 0:
            nvcc_errors.append(c)
            print(f"Error: failed to compile CUDA kernel {kernel}")
            print(" using command:")
            print(f" {shlex.join(nvcc_cmd)}")
            print(stdout or "")
            print(stderr or "")

    nvcc_threads = []
    for cmd in nvcc_commands:
        from threading import Thread
        t = Thread(target=nvcc_compile, args=(cmd,))
        t.start()
        nvcc_threads.append(t)
    for t in nvcc_threads:
        if nvcc_errors:
            sys.exit(1)
        t.join()
    return len(nvcc_errors)


if __name__ == "__main__":
    r = main(sys.argv)
    sys.exit(r)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import re
from time import monotonic_ns
from collections.abc import Callable
from typing import Any

from xpra.x11.server.core import X11ServerCore
from xpra.net.compression import Compressed
from xpra.util.system import is_Wayland, get_loaded_kernel_modules
from xpra.util.objects import AdHocStruct, merge_dicts
from xpra.util.env import envint, envbool
from xpra.common import NotificationID
from xpra.server.shadow.root_window_model import RootWindowModel
from xpra.server.shadow.gtk_shadow_server_base import GTKShadowServerBase
from xpra.server.shadow.gtk_root_window_model import GTKImageCapture
from xpra.server.shadow.shadow_server_base import ShadowServerBase, try_setup_capture
from xpra.x11.server.server_uuid import del_mode, del_uuid
from xpra.x11.gtk.prop import prop_get
from xpra.x11.bindings.window import X11WindowBindings
from xpra.gtk.util import get_default_root_window, get_root_size
from xpra.gtk.error import xsync, xlog
from xpra.log import Logger

log = Logger("x11", "shadow")

XSHM: bool = envbool("XPRA_SHADOW_XSHM", True)
POLL_CURSOR: int = envint("XPRA_SHADOW_POLL_CURSOR", 20)
NVFBC: bool = envbool("XPRA_SHADOW_NVFBC", True)
GSTREAMER: bool = envbool("XPRA_SHADOW_GSTREAMER", False)
nvfbc = None
if NVFBC:
    try:
        from xpra.codecs.nvidia.nvfbc.capture import get_capture_module, get_capture_instance

        nvfbc = get_capture_module()
        if nvfbc:
            nvfbc.init_nvfbc_library()
    except Exception:
        log("NvFBC Capture is not available", exc_info=True)
        NVFBC = False
        nvfbc = None


def window_matches(wspec, model_class):
    wspec = list(wspec)
    try:
        wspec.remove("skip-children")
    except ValueError:
        skip_children = False
    else:
        skip_children = True
    wb = X11WindowBindings()
    with xsync:
        allw = [wxid for wxid in wb.get_all_x11_windows() if
                not wb.is_inputonly(wxid) and wb.is_mapped(wxid)]
        names = {}
        commands = {}
        classes = {}
        for wxid in allw:
            name = prop_get(wxid, "_NET_WM_NAME", "utf8", True) or prop_get(wxid, "WM_NAME", "latin1", True)
            if name:
                names[wxid] = name
            command = prop_get(wxid, "WM_COMMAND", "latin1", True)
            if command:
                commands[wxid] = command.strip("\0")
            class_instance = wb.getClassHint(wxid)
            if class_instance:
                classes[wxid] = class_instance[0].decode("latin1")

        def matchre(re_str, xid_dict):
            xids = []
            try:
                re_c = re.compile(re_str, re.IGNORECASE)
            except re.error:
                log.error("Error: invalid window regular expression %r", re_str)
            else:
                for wxid, vstr in xid_dict.items():
                    if re_c.match(vstr):
                        xids.append(wxid)
            return xids

        def i(v):
            try:
                if v.startswith("0x"):
                    return int(v, 16)
                return int(v)
            except ValueError:
                return 0

        windows = []
        skip = []
        for m in wspec:
            xids = []
            if m.startswith("xid="):
                m = m[4:]
            xid = i(m)
            if xid:
                xids.append(xid)
            elif m.startswith("pid="):
                pid = i(m[4:])
                if pid:
                    try:
                        from xpra.x11.bindings.res import ResBindings  # pylint: disable=import-outside-toplevel
                    except ImportError:
                        XRes = None
                    else:
                        XRes = ResBindings()
                    if XRes and XRes.check_xres():
                        for xid in names:
                            if XRes.get_pid(xid) == pid:
                                xids.append(xid)
            elif m.startswith("command="):
                command = m[len("command="):]
                xids += matchre(command, commands)
            elif m.startswith("class="):
                _class = m[len("class="):]
                xids += matchre(_class, classes)
            else:
                # assume this is a window name:
                xids += matchre(m, names)
            for xid in sorted(xids):
                if xid in skip:
                    continue
                # log.info("added %s", hex(xid))
                windows.append(xid)
                if skip_children:
                    children = wb.get_all_children(xid)
                    skip += children
                # for cxid in wb.get_all_children(xid):
                #    if cxid not in windows:
                #        windows.append(cxid)
        models = {}
        for xid in windows:
            geom = wb.getGeometry(xid)
            if not geom:
                continue
            x, y, w, h = geom[:4]
            # absp = wb.get_absolute_position(xid)
            if w > 0 and h > 0:
                title = names.get(xid, "unknown window")
                model = model_class(title, (x, y, w, h))
                models[xid] = model
        log("window_matches(%s, %s)=%s", wspec, model_class, models)
        # find relative position and 'transient-for':
        for xid, model in models.items():
            model.xid = xid
            model.override_redirect = wb.is_override_redirect(xid)
            transient_for_xid = prop_get(xid, "WM_TRANSIENT_FOR", "window", True)
            model.transient_for = None
            if transient_for_xid:
                try:
                    from xpra.x11.gtk.bindings import get_pywindow
                    model.transient_for = get_pywindow(transient_for_xid)
                except ImportError:
                    pass
            rel_parent = model.transient_for
            if not rel_parent:
                parent = xid
                rel_parent = None
                while parent > 0:
                    parent = wb.getParent(parent)
                    rel_parent = models.get(parent)
                    if rel_parent:
                        log.warn(f"Warning: {rel_parent} is the parent of {model}")
                        break
            model.parent = rel_parent
            # "class-instance", "client-machine", "window-type",
            if rel_parent:
                parent_g = rel_parent.get_geometry()
                dx = model.geometry[0] - parent_g[0]
                dy = model.geometry[1] - parent_g[1]
                model.relative_position = dx, dy
                log("relative_position=%s", model.relative_position)
        log("window_matches%s models=%s", (wspec, model_class), models)
        return models.values()


class XImageCapture:
    __slots__ = ("xshm", "xwindow", "XImage")

    def __init__(self, xwindow: int):
        log("XImageCapture(%#x)", xwindow)
        self.xshm = None
        self.xwindow = xwindow
        from xpra.x11.bindings.ximage import XImageBindings  # pylint: disable=import-outside-toplevel
        self.XImage = XImageBindings()
        assert XSHM and self.XImage.has_XShm(), "no XShm support"
        if is_Wayland():
            log.warn("Warning: shadow servers do not support Wayland")
            log.warn(" please switch to X11 for shadow support")

    def __repr__(self):
        return f"XImageCapture({self.xwindow:x})"

    def get_type(self) -> str:
        return "XImageCapture"

    def clean(self) -> None:
        self.close_xshm()

    def close_xshm(self) -> None:
        xshm = self.xshm
        if self.xshm:
            self.xshm = None
            with xlog:
                xshm.cleanup()

    def _err(self, e, op="capture pixels") -> None:
        if getattr(e, "msg", None) == "BadMatch":
            log("BadMatch - temporary error in %s of window #%x", op, self.xwindow, exc_info=True)
        else:
            log.warn("Warning: failed to %s of window %#x:", op, self.xwindow)
            log.warn(" %s", e)
        self.close_xshm()

    def refresh(self) -> bool:
        if self.xshm:
            # discard to ensure we will call XShmGetImage next time around
            self.xshm.discard()
            return True
        try:
            with xsync:
                log("%s.refresh() xshm=%s", self, self.xshm)
                self.xshm = self.XImage.get_XShmWrapper(self.xwindow)
                self.xshm.setup()
        except Exception as e:
            self.xshm = None
            self._err(e, "xshm setup")
        return True

    def get_image(self, x: int, y: int, width: int, height: int):
        log("XImageCapture.get_image%s for %#x", (x, y, width, height), self.xwindow)
        if self.xshm is None:
            log("no xshm, cannot get image")
            return None
        start = monotonic_ns()
        try:
            with xsync:
                log("X11 shadow get_image, xshm=%s", self.xshm)
                image = self.xshm.get_image(self.xwindow, x, y, width, height)
                return image
        except Exception as e:
            self._err(e)
            return None
        finally:
            end = monotonic_ns()
            log("X11 shadow captured %s pixels at %i MPixels/s using %s",
                width * height, (width * height / (end - start)), ["GTK", "XSHM"][XSHM])


def setup_nvfbc_capture(window):
    if not NVFBC:
        return None
    ww, wh = window.get_geometry()[2:4]
    capture = get_capture_instance()
    capture.init_context(ww, wh)
    capture.refresh()
    image = capture.get_image(0, 0, ww, wh)
    assert image, "test capture failed"
    return capture


def setup_gstreamer_capture(window):
    if not GSTREAMER:
        return None
    xid = window.get_xid()
    ww, wh = window.get_geometry()[2:4]
    from xpra.codecs.gstreamer.capture import Capture
    el = "ximagesrc"
    if xid >= 0:
        el += f" xid={xid} startx=0 starty=0"
    if ww > 0:
        el += f" endx={ww}"
    if wh > 0:
        el += f" endy={wh}"
    capture = Capture(el, width=ww, height=wh)
    capture.start()
    image = capture.get_image(0, 0, ww, wh)
    if not image:
        log("gstreamer capture failed to return an image")
        return None
    return capture


def setup_xshm_capture(window):
    if not XSHM:
        return None
    xid = window.get_xid()
    try:
        from xpra.x11.bindings.ximage import XImageBindings  # pylint: disable=import-outside-toplevel
        XImage = XImageBindings()
    except ImportError as e:
        log(f"not using X11 capture using bindings: {e}")
        return None
    if XImage.has_XShm():
        return XImageCapture(xid)
    return None


def setup_gtk_capture(window):
    return GTKImageCapture(window)


CAPTURE_BACKENDS: dict[str, Callable] = {
    "nvfbc": setup_nvfbc_capture,
    "gstreamer": setup_gstreamer_capture,
    "xshm": setup_xshm_capture,
    "x11": setup_xshm_capture,
    "gtk": setup_gtk_capture,
}


class X11ShadowModel(RootWindowModel):
    __slots__ = ("xid", "override_redirect", "transient_for", "parent", "relative_position")

    def __init__(self, root_window, capture=None, title="", geometry=None):
        super().__init__(root_window, capture, title, geometry)
        self.property_names += ["transient-for", "parent", "relative-position"]
        self.dynamic_property_names += ["transient-for", "parent", "relative-position"]
        self.override_redirect: bool = False
        self.transient_for = None
        self.parent = None
        self.relative_position = ()
        try:
            self.xid = root_window.get_xid()
            self.property_names.append("xid")
        except Exception:
            self.xid = 0

    def get_id(self) -> int:
        return self.xid

    def __repr__(self) -> str:
        info = ", OR" if self.override_redirect else ""
        return f"X11ShadowModel({self.capture} : {self.geometry} : {self.xid:x}{info})"


# FIXME: warning: this class inherits from ServerBase twice..
# so many calls will happen twice there (__init__ and init)
class ShadowX11Server(GTKShadowServerBase, X11ServerCore):

    def __init__(self, attrs: dict[str, str]):
        GTKShadowServerBase.__init__(self, attrs)
        X11ServerCore.__init__(self)
        self.session_type = "X11"
        self.modify_keymap = False
        self.backend = attrs.get("backend", "x11")

    def get_server_mode(self) -> str:
        return "X11 shadow"

    def init(self, opts) -> None:
        GTKShadowServerBase.init(self, opts)
        # don't call init on X11ServerCore,
        # this would call up to GTKServerBase.init(opts) again:
        X11ServerCore.do_init(self, opts)
        self.modify_keymap = opts.keyboard_layout.lower() in ("client", "auto")

    def set_keymap(self, server_source, force: bool = False) -> None:
        if self.readonly:
            return
        if self.modify_keymap:
            X11ServerCore.set_keymap(self, server_source, force)
        else:
            ShadowServerBase.set_keymap(self, server_source, force)

    def cleanup(self) -> None:
        GTKShadowServerBase.cleanup(self)
        X11ServerCore.cleanup(self)
        for fn in (del_mode, del_uuid):
            try:
                fn()
            except Exception:
                log("cleanup() failed to remove X11 attribute", exc_info=True)
        self.do_clean_session_files("xauthority")

    def setup_capture(self):
        capture = try_setup_capture(CAPTURE_BACKENDS, self.backend, self.root)
        log(f"setup_capture() {self.backend} - {self.root}: {capture}")
        return capture

    def get_root_window_model_class(self) -> type:
        return X11ShadowModel

    def makeDynamicWindowModels(self):
        assert self.window_matches
        rwmc = self.get_root_window_model_class()
        root = get_default_root_window()

        def model_class(title, geometry):
            model = rwmc(root, self.capture, title, geometry)
            model.dynamic_property_names.append("size-hints")
            return model

        return window_matches(self.window_matches, model_class)

    def client_startup_complete(self, ss) -> None:
        super().client_startup_complete(ss)
        log("is_Wayland()=%s", is_Wayland())
        if is_Wayland():
            ss.may_notify(NotificationID.SHADOWWAYLAND,
                          "Wayland Shadow Server",
                          "This shadow session seems to be running under wayland,\n"
                          "the screen scraping will probably come up empty",
                          icon_name="unticked")

    def last_client_exited(self) -> None:
        GTKShadowServerBase.last_client_exited(self)
        X11ServerCore.last_client_exited(self)

    def do_get_cursor_data(self) -> tuple[Any, Any]:
        return X11ServerCore.get_cursor_data(self)

    def send_initial_data(self, ss, c, send_ui: bool, share_count: int) -> None:
        super().send_initial_data(ss, c, send_ui, share_count)
        if getattr(ss, "ui_client", True) and getattr(ss, "send_windows", True):
            self.verify_capture(ss)

    def verify_capture(self, ss) -> None:
        log(f"verify_capture({ss})")
        nid = NotificationID.DISPLAY
        try:
            capture = GTKImageCapture(self.root)
            bdata = capture.take_screenshot()[-1]
            title = body = ""
            if any(b != 0 for b in bdata):
                log("verify_capture(%s) succeeded", ss)
            else:
                log.warn("Warning: shadow screen capture is blank")
                body = "The shadow display capture is blank"
                if get_loaded_kernel_modules("vboxguest", "vboxvideo"):
                    body += "\nthis may be caused by the VirtualBox video driver."
                if is_Wayland():
                    body += "Wayland sessions cannot be shadowed in X11 mode."
                title = "Shadow Capture Failure"
            log("verify_capture: title=%r, body=%r", title, body)
            if title and body:
                ss.may_notify(nid, title, body, icon_name="server")
        except Exception as e:
            ss.may_notify(nid, "Shadow Error", f"Error shadowing the display:\n{e}", icon_name="bugs")

    def make_hello(self, source) -> dict[str, Any]:
        capabilities = X11ServerCore.make_hello(self, source)
        capabilities.update(GTKShadowServerBase.make_hello(self, source))
        capabilities["server_type"] = "X11 Shadow"
        return capabilities

    def get_info(self, proto, *_args) -> dict[str, Any]:
        info = X11ServerCore.get_info(self, proto)
        merge_dicts(info, ShadowServerBase.get_info(self, proto))
        info.setdefault("features", {})["shadow"] = True
        info.setdefault("server", {})["type"] = "Python/bindings/x11-shadow"
        return info

    def do_make_screenshot_packet(self) -> tuple[str, int, int, str, int, Compressed]:
        capture = GTKImageCapture(self.root)
        w, h, encoding, rowstride, data = capture.take_screenshot()
        assert encoding == "png"  # use fixed encoding for now
        # pylint: disable=import-outside-toplevel
        return "screenshot", w, h, encoding, rowstride, Compressed(encoding, data)


def snapshot(filename) -> int:
    # pylint: disable=import-outside-toplevel
    from io import BytesIO
    from xpra.util.str_fn import memoryview_to_bytes
    root = get_default_root_window()
    capture = try_setup_capture(CAPTURE_BACKENDS, "auto", root)
    capture.refresh()
    w, h = get_root_size()
    image = capture.get_image(0, 0, w, h)
    log(f"snapshot: {capture.get_image}(0, 0, {w}, {h})={image}")
    from PIL import Image
    fmt = image.get_pixel_format().replace("X", "A")
    pixels = memoryview_to_bytes(image.get_pixels())
    log(f"converting {len(pixels)} bytes in format {fmt} to RGBA")
    if len(fmt) == 3:
        target = "RGB"
    else:
        target = "RGBA"
    pil_image = Image.frombuffer(target, (w, h), pixels, "raw", fmt, image.get_rowstride())
    if target != "RGB":
        pil_image = pil_image.convert("RGB")
    buf = BytesIO()
    pil_image.save(buf, "png")
    data = buf.getvalue()
    buf.close()
    with open(filename, "wb") as f:
        f.write(data)
    return 0


def main(*args) -> int:
    assert len(args) > 0
    if args[0].endswith(".png"):
        return snapshot(args[0])

    def cb(title, geom):
        s = AdHocStruct()
        s.title = title
        s.geometry = geom
        return s

    from xpra.x11.gtk import gdk_display_source  # pylint: disable=import-outside-toplevel, no-name-in-module
    gdk_display_source.init_gdk_display_source()  # @UndefinedVariable
    for w in window_matches(args, cb):
        print(f"{w}")
    return 0


if __name__ == "__main__":
    import sys

    if len(sys.argv) == 1:
        cmd = sys.argv[0]
        print(f"usage: {cmd} filename.png")
        print(f"usage: {cmd} windowname|windowpid")
        r = 1
    else:
        r = main(*sys.argv[1:])
    sys.exit(r)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2009 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os.path

from xpra.os_util import gi_import
from xpra.util.version import XPRA_VERSION
from xpra.scripts.config import get_build_info
from xpra.gtk.window import add_close_accel
from xpra.log import Logger

Gtk = gi_import("Gtk")

log = Logger("info")

APPLICATION_NAME = "Xpra"
SITE_DOMAIN = "xpra.org"
SITE_URL = f"https://{SITE_DOMAIN}/"

GPL2 = ""

MAIN_AUTHORS = (
    'Antoine Martin <antoine@xpra.org>',
    'Nathaniel Smith <njs@pobox.com>',
    'Serviware - Arthur Huillet <ahuillet@serviware.com>',
)


def load_license():
    global GPL2
    if not GPL2:
        from xpra.platform.paths import get_resources_dir  # pylint: disable=import-outside-toplevel
        gpl2_file = os.path.join(get_resources_dir(), "COPYING")
        if os.path.exists(gpl2_file):
            with open(gpl2_file, mode="rb") as f:
                GPL2 = f.read().decode("latin1")
    return GPL2 or "Your installation may be corrupted," \
                   " the license text for GPL version 2 could not be found," \
                   "\nplease refer to:\nhttp://www.gnu.org/licenses/gpl-2.0.txt"


about_dialog = None


def close_about(*_args):
    if about_dialog:
        about_dialog.hide()


def about(on_close=close_about, parent: Gtk.Window | None = None):
    global about_dialog
    if about_dialog:
        about_dialog.show()
        about_dialog.present()
        return
    from xpra.platform.paths import get_icon  # pylint: disable=import-outside-toplevel
    xpra_icon = get_icon("xpra.png")
    dialog = Gtk.AboutDialog()
    dialog.set_name("Xpra")
    dialog.set_version(XPRA_VERSION)
    if parent:
        dialog.set_transient_for(parent)
        dialog.set_destroy_with_parent(True)
    dialog.set_authors(MAIN_AUTHORS)
    _license = load_license()
    dialog.set_license(_license)
    dialog.set_comments("\n".join(get_build_info()))
    dialog.set_website(SITE_URL)
    dialog.set_website_label(SITE_DOMAIN)
    if xpra_icon:
        dialog.set_logo(xpra_icon)
    if hasattr(dialog, "set_program_name"):
        dialog.set_program_name(APPLICATION_NAME)
    dialog.connect("response", on_close)
    add_close_accel(dialog, on_close)
    about_dialog = dialog
    dialog.show()


def main() -> int:
    # pylint: disable=import-outside-toplevel
    from xpra.platform import program_context
    from xpra.platform.gui import init as gui_init
    with program_context("About"):
        gui_init()
    about(on_close=Gtk.main_quit)
    Gtk.main()
    return 0


if __name__ == "__main__":
    main()



# This file is part of Xpra.
# This file is based on websockify/websocket.py from the websockify project
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2011 Joel Martin
# Copyright (C) 2016 Pierre Ossman
# Licensed under LGPL version 3 (see docs/LICENSE.LGPL-3)

import struct

from xpra.common import SizedBuffer
from xpra.net.websockets.common import OPCODE
from xpra.net.websockets.mask import hybi_unmask


def close_packet(code: int = 1000, reason: str = "") -> bytes:
    data = struct.pack("!H", code)
    if reason:
        # should validate that encoded data length is less than 125, meh
        data += reason.encode("utf-8")
    header = encode_hybi_header(OPCODE.CLOSE, len(data), has_mask=False, fin=True)
    return header + data


def encode_hybi_header(opcode, payload_len, has_mask=False, fin=True) -> bytes:
    """ Encode a HyBi style WebSocket frame """
    if (opcode & 0x0f) != opcode:
        raise ValueError(f"invalid opcode {opcode:x}")
    mask_bit = 0x80 * has_mask
    b1 = opcode | (0x80 * fin)
    if payload_len <= 125:
        return struct.pack('>BB', b1, payload_len | mask_bit)
    if payload_len < 65536:
        return struct.pack('>BBH', b1, 126 | mask_bit, payload_len)
    return struct.pack('>BBQ', b1, 127 | mask_bit, payload_len)


def decode_hybi(buf: SizedBuffer) -> tuple[int, SizedBuffer, int, int] | None:
    """ Decode HyBi style WebSocket packets """
    blen = len(buf)
    hlen = 2
    if blen < hlen:
        # log("decode_hybi_header() buffer too small: %i", blen)
        return None

    b1, b2 = struct.unpack(">BB", buf[:2])
    opcode = b1 & 0x0f
    fin = bool(b1 & 0x80)
    masked = bool(b2 & 0x80)
    if masked:
        hlen += 4
        if blen < hlen:
            # log("decode_hybi_header() buffer too small for mask: %i", blen)
            return None

    payload_len = b2 & 0x7f
    if payload_len == 126:
        hlen += 2
        if blen < hlen:
            # log("decode_hybi_header() buffer too small for 126 payload: %i", blen)
            return None
        payload_len = struct.unpack('>H', buf[2:4])[0]
    elif payload_len == 127:
        hlen += 8
        if blen < hlen:
            # log("decode_hybi_header() buffer too small for 127 payload: %i", blen)
            return None
        payload_len = struct.unpack('>Q', buf[2:10])[0]

    # log("decode_hybi_header() decoded header '%s': hlen=%i,
    #    payload_len=%i, buffer len=%i", binascii.hexlify(buf[:hlen]), hlen, payload_len, blen)
    length = hlen + payload_len
    if blen < length:
        # log("decode_hybi_header() buffer too small for payload: %i (needed %i)", blen, length)
        return None

    if masked:
        payload = hybi_unmask(buf, hlen - 4, payload_len)
    else:
        payload = buf[hlen:length]
    # log("decode_hybi_header() payload_len=%i, hlen=%i,
    #    length=%i, fin=%s", payload_len, hlen, length, fin)
    return opcode, payload, length, fin



# This file is part of Xpra.
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Final

from xpra.x11.bindings.window import constants, X11WindowBindings
from xpra.log import Logger

log = Logger("x11", "focus")

X11Window = X11WindowBindings()

CurrentTime: Final[int] = constants["CurrentTime"]
SubstructureNotifyMask: Final[int] = constants["SubstructureNotifyMask"]
SubstructureRedirectMask: Final[int] = constants["SubstructureRedirectMask"]


def send_wm_take_focus(xid: int, timestamp: int = CurrentTime) -> None:
    log("sending WM_TAKE_FOCUS: %#x, X11 timestamp=%r", xid, int(timestamp or 0))
    if timestamp < 0:
        timestamp = 0
    elif timestamp > 0xFFFFFFFF:
        raise OverflowError(f"invalid time: {timestamp:x}")
    elif timestamp > 0x7FFFFFFF:
        timestamp = int(0x100000000 - timestamp)
        if timestamp >= 0x80000000:
            timestamp -= 0x80000000
    X11Window.sendClientMessage(xid, xid, False, 0,
                                "WM_PROTOCOLS",
                                "WM_TAKE_FOCUS", timestamp)


def send_wm_delete_window(xid: int, timestamp: int = CurrentTime) -> None:
    log("sending WM_DELETE_WINDOW to %#x", xid)
    X11Window.sendClientMessage(xid, xid, False, 0,
                                "WM_PROTOCOLS",
                                "WM_DELETE_WINDOW",
                                timestamp)


def send_wm_workspace(root_xid: int, xid: int, workspace: int = 0, timestamp: int = CurrentTime) -> None:
    event_mask = SubstructureNotifyMask | SubstructureRedirectMask
    X11Window.sendClientMessage(root_xid, xid, False, event_mask,
                                "_NET_WM_DESKTOP",
                                workspace,
                                timestamp)


def send_wm_request_frame_extents(root_xid: int, xid: int, timestamp: int = CurrentTime) -> None:
    event_mask = SubstructureNotifyMask | SubstructureRedirectMask
    X11Window.sendClientMessage(root_xid, xid, False, event_mask,
                                "_NET_REQUEST_FRAME_EXTENTS",
                                0,
                                timestamp)



# This file is part of Xpra.
# Copyright (C) 2014 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.common import Self


class WindowBorder:

    def __init__(self, shown: bool = True,
                 red: float = 0.9, green: float = 0.1, blue: float = 0.1, alpha: float = 0.6, size: int = 4):
        self.shown: bool = shown
        self.red: float = red
        self.green: float = green
        self.blue: float = blue
        self.alpha: float = alpha
        self.size: int = size

    def toggle(self) -> None:
        self.shown = not self.shown

    def clone(self) -> Self:
        return WindowBorder(self.shown, self.red, self.green, self.blue, self.alpha, self.size)

    def __repr__(self):
        def hex2(v):
            b = int(max(0, min(255, v * 256)))
            if b < 16:
                return f"0{b:X}"
            return f"{b:X}"

        return "WindowBorder(%s, 0x%s%s%s, %s, %s)" % (
            self.shown,
            hex2(self.red), hex2(self.green), hex2(self.blue), self.alpha, self.size,
        )



# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from ctypes import WinDLL, windll, c_int, byref, Structure, POINTER  # @UnresolvedImport
from ctypes.wintypes import DWORD, PDWORD, PBOOL, BOOL, UINT

from xpra.log import Logger
from xpra.util.env import envbool
from xpra.os_util import gi_import

Gtk = gi_import("Gtk")
Gdk = gi_import("Gdk")

log = Logger("win32")

TITLEBAR_TRANSPARENCY = envbool("XPRA_TITLEBAR_TRANSPARENCY", False)

COLORREF = DWORD  # 0x00bbggrr


class COLORIZATIONPARAMS(Structure):
    _fields_ = (
        ('Color', COLORREF),
        ('Afterglow', COLORREF),
        ('ColorBalance', UINT),
        ('AfterglowBalance', UINT),
        ('BlurBalance', UINT),
        ('GlassReflectionIntensity', UINT),
        ('Opaque', BOOL),
    )


PCOLORIZATIONPARAMS = POINTER(COLORIZATIONPARAMS)

try:
    DwmGetColorizationParameters = windll.dwmapi[127]
except Exception:
    DwmGetColorizationParameters = None
else:
    DwmGetColorizationParameters.argyptes = [PCOLORIZATIONPARAMS]
    DwmGetColorizationParameters.restype = c_int


def rgba(c):
    r, g, b = (c >> 16) & 0xFF, (c >> 8) & 0xFF, c & 0xFF
    a = (c >> 24) & 0xff
    return r, g, b, a


def get_frame_color():
    if not DwmGetColorizationParameters:
        return None
    params = COLORIZATIONPARAMS()
    r = DwmGetColorizationParameters(byref(params))
    log("DwmGetColorizationParameters(..)=%i", r)
    if r:
        return None
    return params.Color


def get_colorization_color():
    dwmapi = WinDLL("dwmapi", use_last_error=True)
    DwmGetColorizationColor = dwmapi.DwmGetColorizationColor
    DwmGetColorizationColor.restype = c_int
    DwmGetColorizationColor.argtypes = [PDWORD, PBOOL]
    color = DWORD()
    opaque = BOOL()
    r = DwmGetColorizationColor(byref(color), byref(opaque))
    log("DwmGetColorizationColor(..)=%i", r)
    if r:
        return None
    return color.value


def match_window_color():
    color = get_frame_color() or get_colorization_color()
    if not color:
        return
    r, g, b, a = rgba(color)
    log("rgba(%#x)=%s", color, (r, g, b, a))
    if TITLEBAR_TRANSPARENCY:
        color_str = "rgba(%i, %i, %i, %.2f)" % (r, g, b, a / 255)
    else:
        color_str = "rgb(%i, %i, %i)" % (r, g, b)
    if min(r, g, b) < 128:
        title_color_str = "white"
    else:
        title_color_str = "black"
    if max(abs(0x99 - c) for c in (r, g, b)) > 64:
        title_unfocused_color_str = "#999999"
    else:
        title_unfocused_color_str = "#333333"
    if max(abs(0xff - c) for c in (r, g, b)) > 64:
        unfocused_color_str = "white"
    else:
        unfocused_color_str = "#666666"
    style_provider = Gtk.CssProvider()
    css_data = """
.titlebar {
 background-color: %s;
 background-image: none;
}
headerbar .title {
 color : %s;
}
headerbar .titlebutton {
 color: %s;
}
.titlebar:backdrop {
 background-color: %s;
}
.titlebar:backdrop .title {
 color : %s;
}
headerbar:backdrop .titlebutton {
 color: %s;
}
""" % (
        color_str, title_color_str, title_color_str,
        unfocused_color_str, title_unfocused_color_str, title_unfocused_color_str,
    )
    log("match_window_color() css=%r", css_data)
    style_provider.load_from_data(css_data.encode("latin1"))
    Gtk.StyleContext.add_provider_for_screen(
        Gdk.Screen.get_default(),
        style_provider,
        Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION + 1
    )



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.


from ctypes.wintypes import HANDLE, DWORD
from ctypes import byref, sizeof, create_string_buffer, cast, c_char, c_void_p, pointer, POINTER
from threading import Thread
from collections.abc import Callable

from xpra.common import noop
from xpra.log import Logger, consume_verbose_argv
from xpra.exit_codes import ExitCode, ExitValue
from xpra.util.env import envbool
from xpra.util.str_fn import strtobytes
from xpra.platform.win32.common import (
    CloseHandle, ERROR_IO_PENDING, FormatMessageSystem,
    GetCurrentProcess,
    SECURITY_ATTRIBUTES,
)
from xpra.platform.win32.namedpipes.common import (
    OVERLAPPED, INFINITE, WAIT_STR, INVALID_HANDLE,
    SECURITY_DESCRIPTOR, TOKEN_USER, TOKEN_PRIMARY_GROUP,
    CreateEventA, CreateNamedPipeA, ConnectNamedPipe,
    WaitForSingleObject, GetLastError,
    SetSecurityDescriptorDacl, SetSecurityDescriptorSacl,
    InitializeSecurityDescriptor,
    OpenProcessToken, GetTokenInformation,
    SetSecurityDescriptorOwner, SetSecurityDescriptorGroup,
    InitializeAcl, GetLengthSid,
    AddAccessAllowedAce, CreateWellKnownSid,
    SID, ACL, ACCESS_ALLOWED_ACE,
)
from xpra.platform.win32.constants import (
    FILE_FLAG_OVERLAPPED, PIPE_ACCESS_DUPLEX, PIPE_READMODE_BYTE,
    PIPE_UNLIMITED_INSTANCES, PIPE_WAIT, PIPE_TYPE_BYTE, NMPWAIT_USE_DEFAULT_WAIT,
    WAIT_TIMEOUT,
    ACL_REVISION,
    SECURITY_DESCRIPTOR_REVISION,
)

log = Logger("network", "named-pipe", "win32")

UNRESTRICTED = envbool("XPRA_NAMED_PIPE_UNRESTRICTED", False)

ERROR_INSUFFICIENT_BUFFER = 122

FILE_ALL_ACCESS = 0x1f01ff
PIPE_ACCEPT_REMOTE_CLIENTS = 0
ERROR_PIPE_CONNECTED = 535

TIMEOUT = 6000
BUFSIZE = 65536

ERROR_ALLOTTED_SPACE_EXCEEDED = 1344
ERROR_INVALID_ACL = 1336
ERROR_INVALID_SID = 1337
ERROR_REVISION_MISMATCH = 1306
ACL_ERRORS = {
    ERROR_ALLOTTED_SPACE_EXCEEDED: "ERROR_ALLOTTED_SPACE_EXCEEDED",
    ERROR_INVALID_ACL: "ERROR_INVALID_ACL",
    ERROR_INVALID_SID: "ERROR_INVALID_SID",
    ERROR_REVISION_MISMATCH: "ERROR_REVISION_MISMATCH",
}

TokenUser = 0x1
TokenPrimaryGroup = 0x5

WinWorldSid = 1
WinLocalSid = 2
WinAnonymousSid = 13

STANDARD_RIGHTS_ALL = 0x001F0000
SPECIFIC_RIGHTS_ALL = 0x0000FFFF
GENERIC_ALL = 0x10000000


class NamedPipeListener(Thread):
    def __init__(self, pipe_name: str, new_connection_cb: Callable = noop):
        log("NamedPipeListener(%s, %s)", pipe_name, new_connection_cb)
        self.pipe_name = pipe_name
        if new_connection_cb != noop:
            self.new_connection_cb = new_connection_cb
        self.exit_loop = False
        super().__init__(name="NamedPipeListener-%s" % pipe_name)
        self.daemon = True
        self.security_attributes: SECURITY_ATTRIBUTES | None = None
        self.security_descriptor: SECURITY_DESCRIPTOR | None = None
        self.token_process = HANDLE()
        cur_proc = GetCurrentProcess()
        log("GetCurrentProcess()=%#x", cur_proc)
        TOKEN_QUERY = 0x8
        if not OpenProcessToken(HANDLE(cur_proc), TOKEN_QUERY, byref(self.token_process)):
            raise OSError()
        log("process=%s", self.token_process.value)

    def __repr__(self):
        return "NamedPipeListener(%r)" % self.pipe_name

    def stop(self) -> None:
        log("%s.stop()", self)
        self.exit_loop = True

    def run(self) -> ExitValue:
        log("%s.run()", self)
        try:
            self.do_run()
        except Exception:
            log.error("Error: named pipe '%s'", self.pipe_name, exc_info=True)
            return ExitCode.FAILURE
        tp = self.token_process
        if tp:
            self.token_process = INVALID_HANDLE
            CloseHandle(tp)
        self.security_attributes = None
        self.security_descriptor = None
        return ExitCode.OK

    def do_run(self) -> None:
        pipe_handle = INVALID_HANDLE
        while not self.exit_loop:
            if pipe_handle == INVALID_HANDLE:
                try:
                    pipe_handle = self.CreatePipeHandle()
                except Exception as e:
                    log("CreatePipeHandle()", exc_info=True)
                    log.error("Error: failed to create named pipe")
                    log.error(" at path '%s'", self.pipe_name)
                    log.estr(e)
                    return
                log("CreatePipeHandle()=%s", pipe_handle)
                if pipe_handle == INVALID_HANDLE:
                    log.error("Error: invalid handle for named pipe '%s'", self.pipe_name)
                    err: int = GetLastError()
                    log.error(" '%s' (%i)", FormatMessageSystem(err).rstrip("\n\r."), err)
                    return
            event = CreateEventA(None, True, False, None)
            overlapped = OVERLAPPED()
            overlapped.hEvent = event
            overlapped.Internal = None
            overlapped.InternalHigh = None
            overlapped.union.Pointer = None
            r = ConnectNamedPipe(pipe_handle, overlapped)
            log("ConnectNamedPipe()=%s", r)
            if self.exit_loop:
                break
            if r == 0:
                err = GetLastError()
                log("GetLastError()=%s (%i)", FormatMessageSystem(err).rstrip("\n\r"), err)
                if err == ERROR_PIPE_CONNECTED:
                    "non-zero, but OK!"
                elif err == ERROR_IO_PENDING:
                    while not self.exit_loop:
                        r = WaitForSingleObject(event, INFINITE)
                        log("WaitForSingleObject(..)=%s", WAIT_STR.get(r, r))
                        if r == WAIT_TIMEOUT:
                            continue
                        if r == 0:
                            break
                        log.error("Error: cannot connect to named pipe '%s'", self.pipe_name)
                        log.error(" %s", WAIT_STR.get(r, r))
                        CloseHandle(pipe_handle)
                        pipe_handle = INVALID_HANDLE
                        break
                else:
                    log.error("Error: cannot connect to named pipe '%s'", self.pipe_name)
                    log.error(" error %s", err)
                    CloseHandle(pipe_handle)
                    pipe_handle = INVALID_HANDLE
                if self.exit_loop:
                    break
            # from now on, the pipe_handle will be managed elsewhere:
            if pipe_handle != INVALID_HANDLE:
                self.new_connection_cb("named-pipe", self, pipe_handle)
                pipe_handle = INVALID_HANDLE
        self.close_handle(pipe_handle)

    def close_handle(self, pipe_handle: HANDLE) -> None:
        log("CloseHandle(%s)", pipe_handle)
        if pipe_handle == INVALID_HANDLE:
            return
        try:
            CloseHandle(pipe_handle)
        except Exception:
            log("CloseHandle(%s)", pipe_handle, exc_info=True)

    def new_connection(self, socktype, listener, pipe_handle: HANDLE) -> None:
        log("new_connection(%s, %s, %s)", socktype, listener, pipe_handle)
        self.close_handle(pipe_handle)

    def CreatePipeHandle(self) -> HANDLE:
        sa = self.CreatePipeSecurityAttributes()
        log("CreateNamedPipeA using %s (UNRESTRICTED=%s)", sa, UNRESTRICTED)
        return CreateNamedPipeA(strtobytes(self.pipe_name), PIPE_ACCESS_DUPLEX | FILE_FLAG_OVERLAPPED,
                                PIPE_TYPE_BYTE | PIPE_READMODE_BYTE | PIPE_WAIT | PIPE_ACCEPT_REMOTE_CLIENTS,
                                PIPE_UNLIMITED_INSTANCES, BUFSIZE, BUFSIZE, NMPWAIT_USE_DEFAULT_WAIT,
                                byref(sa))

    def GetToken(self, token_type, token_struct):
        assert self.token_process
        data_size = DWORD()
        if not GetTokenInformation(self.token_process, token_type, 0, 0, byref(data_size)):
            if GetLastError() != ERROR_INSUFFICIENT_BUFFER:
                raise OSError()
        log("GetTokenInformation data size %#x", data_size.value)
        # noinspection PyTypeChecker
        buftype = c_char * (data_size.value + 1)
        token_data = buftype()
        if not GetTokenInformation(self.token_process, token_type, byref(token_data), data_size.value,
                                   byref(data_size)):
            raise OSError()
        token = cast(token_data, POINTER(token_struct)).contents
        return token

    def CreatePipeSecurityAttributes(self) -> SECURITY_ATTRIBUTES:
        user = self.GetToken(TokenUser, TOKEN_USER)
        user_SID = user.SID.contents
        log("user SID=%s, attributes=%#x", user_SID, user.ATTRIBUTES)

        group = self.GetToken(TokenPrimaryGroup, TOKEN_PRIMARY_GROUP)
        group_SID = group.PrimaryGroup.contents
        log("group SID=%s", group_SID)

        SD = SECURITY_DESCRIPTOR()
        self.security_descriptor = SD
        log("SECURITY_DESCRIPTOR=%s", SD)
        if not InitializeSecurityDescriptor(byref(SD), SECURITY_DESCRIPTOR_REVISION):
            raise OSError()  # @UndefinedVariable
        log("InitializeSecurityDescriptor: %s", SD)
        if not SetSecurityDescriptorOwner(byref(SD), user.SID, False):
            raise OSError()
        log("SetSecurityDescriptorOwner: %s", SD)
        if not SetSecurityDescriptorGroup(byref(SD), group.PrimaryGroup, False):
            raise OSError()
        log("SetSecurityDescriptorGroup: %s", SD)
        SA = SECURITY_ATTRIBUTES()
        log("CreatePipeSecurityObject() SECURITY_ATTRIBUTES=%s", SA)
        if not UNRESTRICTED:
            SA.descriptor = SD
            SA.bInheritHandle = False
            return SA
        if not SetSecurityDescriptorSacl(byref(SD), False, None, False):
            raise OSError()
        if not SetSecurityDescriptorDacl(byref(SD), True, None, False):
            raise OSError()
        # this doesn't work - and I don't know why:
        # SECURITY_NT_AUTHORITY = 5
        # sia_anonymous = SID_IDENTIFIER_AUTHORITY((0, 0, 0, 0, 0, SECURITY_NT_AUTHORITY))
        # log("SID_IDENTIFIER_AUTHORITY(SECURITY_NT_AUTHORITY)=%s", sia_anonymous)
        # sid_allow = SID()
        # log("empty SID: %s", sid_allow)
        # if not AllocateAndInitializeSid(byref(sia_anonymous), 1,
        #                         SECURITY_ANONYMOUS_LOGON_RID, 0, 0, 0, 0, 0, 0, 0,
        #                         byref(sid_allow),
        #                         ):
        #    raise WindowsError()
        #    log("AllocateAndInitializeSid(..) sid_anonymous=%s", sid_allow)
        sid_allow = SID()
        sid_size = DWORD(sizeof(SID))
        sid_type = WinWorldSid
        SECURITY_MAX_SID_SIZE = 68
        assert sizeof(SID) >= SECURITY_MAX_SID_SIZE
        if not CreateWellKnownSid(sid_type, None, byref(sid_allow), byref(sid_size)):
            log.error("error=%s", GetLastError())
            raise OSError()
        assert sid_size.value <= SECURITY_MAX_SID_SIZE
        log("CreateWellKnownSid(..) sid_allow=%s, sid_size=%s", sid_allow, sid_size)

        acl_size = sizeof(ACL)
        acl_size += 2 * (sizeof(ACCESS_ALLOWED_ACE) - sizeof(DWORD))
        acl_size += GetLengthSid(byref(sid_allow))
        acl_size += GetLengthSid(byref(user.SID.contents))
        # acl_size += GetLengthSid(user.SID)
        acl_data = create_string_buffer(acl_size)
        acl = cast(acl_data, POINTER(ACL)).contents
        log("acl_size=%s, acl_data=%s, acl=%s", acl_size, acl_data, acl)
        if not InitializeAcl(byref(acl), acl_size, ACL_REVISION):
            raise OSError()
        log("InitializeAcl(..) acl=%s", acl)

        rights = STANDARD_RIGHTS_ALL | SPECIFIC_RIGHTS_ALL
        add_sid = user.SID
        r = AddAccessAllowedAce(byref(acl), ACL_REVISION, rights, add_sid)
        if r == 0:
            err = GetLastError()
            log("AddAccessAllowedAce(..)=%s", ACL_ERRORS.get(err, err))
            raise OSError()

        rights = STANDARD_RIGHTS_ALL | SPECIFIC_RIGHTS_ALL
        add_sid = byref(sid_allow)
        r = AddAccessAllowedAce(byref(acl), ACL_REVISION, rights, add_sid)
        if r == 0:
            err = GetLastError()
            log("AddAccessAllowedAce(..)=%s", ACL_ERRORS.get(err, err))
            raise OSError()
        if not SetSecurityDescriptorDacl(byref(SD), True, byref(acl), False):
            raise OSError()
        SA.nLength = sizeof(SECURITY_ATTRIBUTES)
        SA.lpSecurityDescriptor = cast(pointer(SD), c_void_p)
        SA.bInheritHandle = True
        self.security_attributes = SA
        return SA


def main():
    import sys
    consume_verbose_argv(sys.argv, "win32")
    pipe_name = "Xpra\\Test"
    if len(sys.argv) > 1:
        pipe_name = sys.argv[1]
    listener = NamedPipeListener("\\\\.\\pipe\\%s" % pipe_name)
    listener.run()


if __name__ == "__main__":
    main()



# This file is part of Xpra.
# Copyright (C) 2012 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from collections import namedtuple

from xpra.client.gtk3.window import ClientWindow
from xpra.gtk.window import set_visual
from xpra.util.objects import typedict
from xpra.util.env import envbool
from xpra.log import Logger

log = Logger("opengl", "window")

Rectangle = namedtuple("Rectangle", ("x", "y", "width", "height"))
DrawEvent = namedtuple("DrawEvent", ("area", ))

MONITOR_REINIT = envbool("XPRA_OPENGL_MONITOR_REINIT", False)


class GLClientWindowBase(ClientWindow):

    def __repr__(self):
        return f"GLClientWindow({self.wid} : {self._backing})"

    def get_backing_class(self) -> type:
        raise NotImplementedError()

    def is_GL(self) -> bool:
        return True

    def spinner(self, ok: bool) -> None:
        b = self._backing
        log("spinner(%s) opengl window %s: backing=%s", ok, self.wid, b)
        if not b:
            return
        b.paint_spinner = self.can_have_spinner() and not ok
        log("spinner(%s) backing=%s, paint_screen=%s, paint_spinner=%s",
            ok, b._backing, b.paint_screen, b.paint_spinner)
        if b._backing and b.paint_screen:
            w, h = self.get_size()
            self.repaint(0, 0, w, h)

    def queue_draw_area(self, x: int, y: int, w: int, h: int) -> None:
        b = self._backing
        if not b:
            return
        b.gl_expose_rect(x, y, w, h)

    def monitor_changed(self, monitor) -> None:
        super().monitor_changed(monitor)
        da = self.drawing_area
        if da and MONITOR_REINIT:
            # re-create the drawing area,
            # which will re-create the opengl context:
            try:
                self.remove(da)
            except Exception:
                log("monitor_changed: failed to remove %s", da)
            self.drawing_area = None
            w, h = self.get_size()
            self.new_backing(w, h)

    def remove_backing(self) -> None:
        b = self._backing
        log("remove_backing() backing=%s", b)
        if b:
            self._backing = None
            b.paint_screen = False
            b.close()
            glarea = b._backing
            log("remove_backing() glarea=%s", glarea)
            if glarea:
                try:
                    self.remove(glarea)
                except Exception:
                    log.warn("Warning: cannot remove %s", glarea, exc_info=True)

    def magic_key(self, *args) -> None:
        b = self._backing
        if self.border:
            self.border.toggle()
            if b:
                with b.gl_context() as ctx:
                    b.gl_init(ctx)
                    b.present_fbo(ctx, 0, 0, *b.size)
                self.repaint(0, 0, *self._size)
        log("gl magic_key%s border=%s, backing=%s", args, self.border, b)

    def set_alpha(self) -> None:
        super().set_alpha()
        rgb_formats = self._client_properties.setdefault("encodings.rgb_formats", [])
        # gl.backing supports BGR(A) too:
        if "RGBA" in rgb_formats:
            rgb_formats.append("BGRA")
        if "RGB" in rgb_formats:
            rgb_formats.append("BGR")

    def do_configure_event(self, event) -> None:
        log("GL do_configure_event(%s)", event)
        ClientWindow.do_configure_event(self, event)
        self._backing.paint_screen = True

    def destroy(self) -> None:
        self.remove_backing()
        super().destroy()

    def init_drawing_area(self) -> None:
        self.drawing_area = None

    def new_backing(self, bw: int, bh: int) -> None:
        widget = super().new_backing(bw, bh)
        if self.drawing_area:
            self.remove(self.drawing_area)
        set_visual(widget, self._has_alpha)
        widget.show()
        self.init_widget_events(widget)
        if self.drawing_area and self.size_constraints:
            # apply min size to the drawing_area:
            thints = typedict(self.size_constraints)
            minsize = thints.intpair("minimum-size", (0, 0))
            self.drawing_area.set_size_request(*minsize)
        self.add(widget)
        self.drawing_area = widget
        # maybe redundant?:
        self.apply_geometry_hints(self.geometry_hints)

    def draw_widget(self, widget, context) -> bool:
        mapped = self.get_mapped()
        backing = self._backing
        log(f"draw_widget({widget}, {context}) {mapped=}, {backing=}", )
        if not mapped:
            return False
        if not backing:
            return False
        return backing.draw_fbo(context)



# This file is part of Xpra.
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.os_util import gi_import
from xpra.x11.bindings.window import X11WindowBindings
from xpra.x11.models.core import CoreX11WindowModel
from xpra.log import Logger

log = Logger("x11", "window", "tray")

GObject = gi_import("GObject")

X11Window = X11WindowBindings()


class TrayGeometryChanged:  # pylint: disable=too-few-public-methods
    __slots__ = ("x", "y", "width", "height")


class SystemTrayWindowModel(CoreX11WindowModel):
    __gproperties__ = CoreX11WindowModel.__common_properties__.copy()
    __gproperties__ |= {
        "tray": (GObject.TYPE_BOOLEAN,
                 "Is the window a system tray icon", "",
                 False,
                 GObject.ParamFlags.READABLE),
    }
    __gsignals__ = CoreX11WindowModel.__common_signals__.copy()
    _property_names = CoreX11WindowModel._property_names + ["tray"]
    # these aren't useful, and could actually cause us problems
    _property_names.remove("opaque-region")
    _property_names.remove("protocols")
    _MODELTYPE = "Tray"

    def __init__(self, xid: int, corral_xid: int):
        super().__init__(xid)
        self.corral_xid: int = corral_xid
        self._updateprop("tray", True)

    def __repr__(self) -> str:
        return f"SystemTrayWindowModel({self.xid:x})"

    def _read_initial_X11_properties(self) -> None:
        self._internal_set_property("has-alpha", True)
        super()._read_initial_X11_properties()

    def move_resize(self, x: int, y: int, width: int, height: int) -> None:
        # Used by clients to tell us where the tray is located on screen
        log("SystemTrayModel.move_resize(%s, %s, %s, %s)", x, y, width, height)
        if not X11Window.MoveResizeWindow(self.corral_xid, x, y, width, height):
            return
        X11Window.MoveResizeWindow(self.xid, 0, 0, width, height)
        self._updateprop("geometry", (x, y, width, height))
        # force a refresh:
        event = TrayGeometryChanged()
        event.x = event.y = 0
        event.width, event.height = self.get_dimensions()
        self.emit("client-contents-changed", event)


GObject.type_register(SystemTrayWindowModel)



# This file is part of Xpra.
# Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# pylint: disable=wrong-import-position

import sys
from time import monotonic
from typing import Any
from collections.abc import Callable

from xpra.util.env import envbool
from xpra.util.version import dict_version_trim
from xpra.os_util import gi_import
from xpra.common import FULL_INFO
from xpra.net.common import PacketType
from xpra.gtk.signals import register_os_signals, register_SIGUSR_signals
from xpra.gtk.keymap import get_default_keymap
from xpra.server import features
from xpra.server.base import ServerBase
from xpra.gtk.util import get_root_size
from xpra.gtk.versions import get_gtk_version_info
from xpra.log import Logger

Gtk = gi_import("Gtk")
Gdk = gi_import("Gdk")
GLib = gi_import("GLib")

UI_THREAD_WATCHER = envbool("XPRA_UI_THREAD_WATCHER")

log = Logger("server", "gtk")
screenlog = Logger("server", "screen")
notifylog = Logger("notify")


class GTKServerBase(ServerBase):
    """
        This is the base class for servers.
        It provides all the generic functions but is not tied
        to a specific backend (X11 or otherwise).
        See X11ServerBase, XpraServer and XpraX11ShadowServer
    """

    def __init__(self):
        log("GTKServerBase.__init__()")
        self.cursor_suspended: bool = False
        self.ui_watcher = None
        self.keymap_changing_timer: int = 0
        super().__init__()

    def watch_keymap_changes(self) -> None:
        # Set up keymap change notification:
        keymap = get_default_keymap()

        # this event can fire many times in succession
        # throttle how many times we call self._keys_changed()

        def keys_changed(*_args):
            if self.keymap_changing_timer:
                return

            def do_keys_changed():
                self.keymap_changing_timer = 0
                self._keys_changed()

            self.keymap_changing_timer = GLib.timeout_add(500, do_keys_changed)

        keymap.connect("keys-changed", keys_changed)

    def stop_keymap_timer(self) -> None:
        kct = self.keymap_changing_timer
        if kct:
            self.keymap_changing_timer = 0
            GLib.source_remove(kct)

    def install_signal_handlers(self, callback: Callable[[int], None]) -> None:
        sstr = self.get_server_mode() + " server"
        register_os_signals(callback, sstr)
        register_SIGUSR_signals(sstr)

    def do_quit(self) -> None:
        log("do_quit: calling Gtk.main_quit")
        Gtk.main_quit()
        log("do_quit: Gtk.main_quit done")
        # from now on, we can't rely on the main loop:
        from xpra.util.system import register_SIGUSR_signals
        register_SIGUSR_signals()

    def late_cleanup(self, stop=True) -> None:
        log("GTKServerBase.late_cleanup(%s)", stop)
        self.stop_keymap_timer()
        super().late_cleanup(stop)
        self.stop_ui_watcher()
        self.close_gtk_display()

    def stop_ui_watcher(self) -> None:
        uiw = self.ui_watcher
        log("stop_ui_watcher() ui watcher=%s", uiw)
        if uiw:
            self.ui_watcher = None
            uiw.stop()

    def close_gtk_display(self) -> None:
        # Close our display(s) first, so the server dying won't kill us.
        # (if gtk has been loaded)
        gdk_mod = sys.modules.get("gi.repository.Gdk")
        # bug 2328: python3 shadow server segfault on Ubuntu 16.04
        # also crashes on Ubuntu 20.04
        close = envbool("XPRA_CLOSE_GTK_DISPLAY", False)
        log("close_gtk_display() close=%s, gdk_mod=%s",
            close, gdk_mod)
        if close and gdk_mod:
            displays = Gdk.DisplayManager.get().list_displays()
            log("close_gtk_display() displays=%s", displays)
            for d in displays:
                log("close_gtk_display() closing %s", d)
                d.close()

    def do_run(self) -> None:
        if UI_THREAD_WATCHER:
            from xpra.platform.ui_thread_watcher import get_UI_watcher  # pylint: disable=import-outside-toplevel
            self.ui_watcher = get_UI_watcher()
            self.ui_watcher.start()
        if features.windows:
            display = Gdk.Display.get_default()
            if display:
                # n = display.get_n_screens()
                # assert n==1, "unsupported number of screens: %i" % n
                screen = display.get_default_screen()
                screen.connect("size-changed", self._screen_size_changed)
                screen.connect("monitors-changed", self._monitors_changed)
        log("do_run() calling %s", Gtk.main)
        Gtk.main()
        log("do_run() end of gtk.main()")

    def make_hello(self, source) -> dict[str, Any]:
        capabilities = super().make_hello(source)
        if "display" in source.wants:
            display = Gdk.Display.get_default()
            if display:
                capabilities |= {
                    "display": display.get_name(),
                }
        if "versions" in source.wants and FULL_INFO >= 2:
            capabilities.setdefault("versions", {}).update(get_gtk_version_info())
        return capabilities

    def get_ui_info(self, proto, *args) -> dict[str, Any]:
        info = super().get_ui_info(proto, *args)
        display = Gdk.Display.get_default()
        if display:
            info.setdefault("server", {}).update(
                {
                    "display": display.get_name(),
                    "root_window_size": self.get_root_window_size(),
                }
            )
        return info

    def suspend_cursor(self, proto) -> None:
        # this is called by shadow and desktop servers
        # when we're receiving pointer events but the pointer
        # is no longer over the active window area,
        # so we have to tell the client to switch back to the default cursor
        if self.cursor_suspended:
            return
        self.cursor_suspended = True
        ss = self.get_server_source(proto)
        if ss:
            ss.cancel_cursor_timer()
            ss.send_empty_cursor()

    def restore_cursor(self, proto) -> None:
        # see suspend_cursor
        if not self.cursor_suspended:
            return
        self.cursor_suspended = False
        ss = self.get_server_source(proto)
        if ss:
            ss.send_cursor()

    def do_get_info(self, proto, *args) -> dict[str, Any]:
        start = monotonic()
        info = super().do_get_info(proto, *args)
        vi = dict_version_trim(get_gtk_version_info())
        vi["type"] = "Python/gtk"
        info.setdefault("server", {}).update(vi)
        log("GTKServerBase.do_get_info took %ims", (monotonic() - start) * 1000)
        return info

    def get_root_window_size(self) -> tuple[int, int]:
        return get_root_size(None)

    def get_max_screen_size(self) -> tuple[int, int]:
        return get_root_size(None)

    def configure_best_screen_size(self) -> tuple[int, int]:
        return self.get_root_window_size()

    def calculate_workarea(self, maxw: int, maxh: int) -> None:
        screenlog("calculate_workarea(%s, %s)", maxw, maxh)
        workarea = Gdk.Rectangle()
        workarea.width = maxw
        workarea.height = maxh
        for ss in self._server_sources.values():
            screen_sizes = ss.screen_sizes
            screenlog("calculate_workarea() screen_sizes(%s)=%s", ss, screen_sizes)
            if not screen_sizes:
                continue
            for display in screen_sizes:
                # avoid error with old/broken clients:
                if not display or not isinstance(display, (list, tuple)):
                    continue
                # display: [':0.0', 2560, 1600, 677, 423, [['DFP2', 0, 0, 2560, 1600, 646, 406]], 0, 0, 2560, 1574]
                if len(display) >= 10:
                    work_x, work_y, work_w, work_h = display[6:10]
                    display_workarea = Gdk.Rectangle()
                    display_workarea.x = work_x
                    display_workarea.y = work_y
                    display_workarea.width = work_w
                    display_workarea.height = work_h
                    screenlog("calculate_workarea() found %s for display %s", display_workarea, display[0])
                    success, workarea = workarea.intersect(display_workarea)
                    if not success:
                        log.warn("Warning: failed to calculate workarea")
                        log.warn(" as intersection of %s and %s", (maxw, maxh), (work_x, work_y, work_w, work_h))
        # sanity checks:
        screenlog("calculate_workarea(%s, %s) workarea=%s", maxw, maxh, workarea)
        max_dim = 32768 - 8192
        if workarea.width == 0 or workarea.height == 0 or workarea.width >= max_dim or workarea.height >= max_dim:
            screenlog.warn("Warning: failed to calculate a common workarea")
            screenlog.warn(f" using the full display area: {maxw}x{maxh}")
            workarea = Gdk.Rectangle()
            workarea.width = maxw
            workarea.height = maxh
        self.set_workarea(workarea)

    def set_workarea(self, workarea):
        """ overridden by seamless servers """

    def set_desktop_geometry(self, width: int, height: int) -> None:
        """ overridden by X11 seamless and desktop servers """

    def _move_pointer(self, device_id: int, wid: int, pos, props=None) -> None:
        x, y = pos
        display = Gdk.Display.get_default()
        display.warp_pointer(display.get_default_screen(), x, y)

    def do_process_button_action(self, *args):
        raise NotImplementedError

    def _process_map_window(self, proto, packet: PacketType) -> None:
        log.info("_process_map_window(%s, %s)", proto, packet)

    def _process_unmap_window(self, proto, packet: PacketType) -> None:
        log.info("_process_unmap_window(%s, %s)", proto, packet)

    def _process_close_window(self, proto, packet: PacketType) -> None:
        log.info("_process_close_window(%s, %s)", proto, packet)

    def _process_configure_window(self, proto, packet: PacketType) -> None:
        log.info("_process_configure_window(%s, %s)", proto, packet)

    def get_notification_icon(self, icon_string: str) -> tuple[str, int, int, bytes] | None:
        try:
            from xpra.notifications.common import get_notification_icon
        except ImportError:
            return None
        return get_notification_icon(icon_string)



# This file is part of Xpra.
# Copyright (C) 2023 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.


EXTENSION_TO_MIMETYPE = {
    ".wasm": "application/wasm",
    ".js": "text/javascript",
    ".css": "text/css",
}



# This file is part of Xpra.
# Copyright (C) 2023 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.


from xpra.codecs.constants import TransientCodecException, CodecStateException

# ctypedef int cudaError_t
# cdef extern from "cuda_runtime_api.h":
#    const char *cudaGetErrorName(cudaError_t err)
#    const char *cudaGetErrorString(cudaError_t err)


def cudacheck(r: int, fn=None) -> None:
    if r == 0:
        return
    msg = get_error_name(r)
    if fn:
        msg = fn + f": {msg!r}"
    if r in TRANSIENT_ERRORS:
        raise TransientCodecException(msg)
    if r in STATE_ERRORS:
        raise CodecStateException(msg)
    raise RuntimeError(msg)


def get_error_name(error: int) -> str:
    return CUDA_ERRORS.get(error) or str(error)


TRANSIENT_ERRORS = (2, 304, 600, 701, 702, 802, 909)

STATE_ERRORS = (3, 4, 200, 201, 202, 205, 206, 207, 208, 210, 211, 216, 219, 400, 401, 708, 709, 712, 713)


CUDA_ERRORS: dict[int, str] = {
    0: "SUCCESS",
    1: "INVALID_VALUE",
    2: "OUT_OF_MEMORY",
    3: "NOT_INITIALIZED",
    4: "DEINITIALIZED",
    5: "PROFILER_DISABLED",
    6: "PROFILER_NOT_INITIALIZED",
    7: "PROFILER_ALREADY_STARTED",
    8: "PROFILER_ALREADY_STOPPED",
    34: "STUB_LIBRARY",
    46: "DEVICE_UNAVAILABLE",
    100: "NO_DEVICE",
    101: "INVALID_DEVICE",
    102: "DEVICE_NOT_LICENSED",
    200: "INVALID_IMAGE",
    201: "INVALID_CONTEXT",
    202: "CONTEXT_ALREADY_CURRENT",
    205: "MAP_FAILED",
    206: "UNMAP_FAILED",
    207: "ARRAY_IS_MAPPED",
    208: "ALREADY_MAPPED",
    209: "NO_BINARY_FOR_GPU",
    210: "ALREADY_ACQUIRED",
    211: "NOT_MAPPED",
    212: "NOT_MAPPED_AS_ARRAY",
    213: "NOT_MAPPED_AS_POINTER",
    214: "ECC_UNCORRECTABLE",
    215: "UNSUPPORTED_LIMIT",
    216: "CONTEXT_ALREADY_IN_USE",
    217: "PEER_ACCESS_UNSUPPORTED",
    218: "INVALID_PTX",
    219: "INVALID_GRAPHICS_CONTEXT",
    220: "NVLINK_UNCORRECTABLE",
    221: "JIT_COMPILER_NOT_FOUND",
    222: "UNSUPPORTED_PTX_VERSION",
    223: "JIT_COMPILATION_DISABLED",
    224: "UNSUPPORTED_EXEC_AFFINITY",
    300: "INVALID_SOURCE",
    301: "FILE_NOT_FOUND",
    302: "SHARED_OBJECT_SYMBOL_NOT_FOUND",
    303: "SHARED_OBJECT_INIT_FAILED",
    304: "OPERATING_SYSTEM",
    400: "INVALID_HANDLE",
    401: "ILLEGAL_STATE",
    500: "NOT_FOUND",
    600: "NOT_READY",
    700: "ILLEGAL_ADDRESS",
    701: "LAUNCH_OUT_OF_RESOURCES",
    702: "LAUNCH_TIMEOUT",
    703: "LAUNCH_INCOMPATIBLE_TEXTURING",
    704: "PEER_ACCESS_ALREADY_ENABLED",
    705: "PEER_ACCESS_NOT_ENABLED",
    708: "PRIMARY_CONTEXT_ACTIVE",
    709: "CONTEXT_IS_DESTROYED",
    710: "ASSERT",
    711: "TOO_MANY_PEERS",
    712: "HOST_MEMORY_ALREADY_REGISTERED",
    713: "HOST_MEMORY_NOT_REGISTERED",
    714: "HARDWARE_STACK_ERROR",
    715: "ILLEGAL_INSTRUCTION",
    716: "MISALIGNED_ADDRESS",
    717: "INVALID_ADDRESS_SPACE",
    718: "INVALID_PC",
    719: "LAUNCH_FAILED",
    720: "COOPERATIVE_LAUNCH_TOO_LARGE",
    800: "NOT_PERMITTED",
    801: "NOT_SUPPORTED",
    999: "UNKNOWN",
    802: "SYSTEM_NOT_READY",
    803: "SYSTEM_DRIVER_MISMATCH",
    804: "COMPAT_NOT_SUPPORTED_ON_DEVICE",
    805: "MPS_CONNECTION_FAILED",
    806: "MPS_RPC_FAILURE",
    807: "MPS_SERVER_NOT_READY",
    808: "MPS_MAX_CLIENTS_REACHED",
    809: "MPS_MAX_CONNECTIONS_REACHED",
    810: "MPS_CLIENT_TERMINATED",
    811: "CDP_NOT_SUPPORTED",
    812: "CDP_VERSION_MISMATCH",
    900: "STREAM_CAPTURE_UNSUPPORTED",
    901: "STREAM_CAPTURE_INVALIDATED",
    902: "STREAM_CAPTURE_MERGE",
    903: "STREAM_CAPTURE_UNMATCHED",
    904: "STREAM_CAPTURE_UNJOINED",
    905: "STREAM_CAPTURE_ISOLATION",
    906: "STREAM_CAPTURE_IMPLICIT",
    907: "CAPTURED_EVENT",
    908: "STREAM_CAPTURE_WRONG_THREAD",
    909: "TIMEOUT",
    910: "GRAPH_EXEC_UPDATE_FAILURE",
    911: "EXTERNAL_DEVICE",
    912: "INVALID_CLUSTER_SIZE",
}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import tempfile
import unittest

from xpra.os_util import WIN32, POSIX
from xpra.util.objects import AdHocStruct
from xpra.util.env import IgnoreWarningsContext
from xpra.scripts.parsing import parse_display_name


def _get_test_socket_dir():
    return tempfile.gettempdir()


def _test_opts():
    socket_dir = _get_test_socket_dir()
    opts = AdHocStruct()
    opts.socket_dirs = [socket_dir]
    opts.socket_dir = socket_dir
    opts.exit_ssh = False
    opts.ssh = "ssh -v "
    opts.remote_xpra = "run-xpra"
    opts.username = ""
    opts.password = ""
    opts.password_file = None
    opts.ssl = "auto"
    opts.ssl_server_hostname = ""
    opts.ssl_cert = ""
    opts.ssl_key = ""
    opts.ssl_ca_certs = "default"
    opts.ssl_ca_data = ""
    opts.ssl_ciphers = "DEFAULT"
    opts.ssl_protocol = "TLSv1_2"
    opts.ssl_client_verify_mode = "optional"
    opts.ssl_server_verify_mode = "required"
    opts.ssl_verify_flags = "X509_STRICT"
    opts.ssl_check_hostname = True
    opts.ssl_server_hostname = ""
    opts.ssl_options = "ALL,NO_COMPRESSION"
    return opts


def _test_parse_display_name(s, e=None):
    opts = _test_opts()

    def err(*args):
        raise Exception(*args)

    r = parse_display_name(err, opts, s)
    if e:
        for k, v in e.items():
            actual = r.get(k)
            if actual != v:
                raise ValueError(f"expected {v!r} but got {actual!r} for {k!r}"
                                 f" from parse_display_name({s!r})={r!r}, expected {e!r}")
    return r


class TestMain(unittest.TestCase):

    def test_parse_display_name(self):
        socket_dir = _get_test_socket_dir()
        if WIN32:
            fd = _test_parse_display_name("named-pipe://FOO")["named-pipe"]
            sd = _test_parse_display_name("FOO")["named-pipe"]
        else:
            fd = _test_parse_display_name("socket:///FOO")
            sd = _test_parse_display_name("/FOO")
        assert sd == fd, f"expected {fd} but got {sd}"
        t = _test_parse_display_name

        def e(s):
            try:
                t(s)
            except Exception:
                pass
            else:
                raise Exception(f"parse_display_name should fail for {s}")

        if POSIX:
            e("ZZZZZZ")
            t("10", {"display_name": ":10", "local": True, "type": "socket"})
            t(socket_dir + "/thesocket", {"display_name": "socket://" + socket_dir + "/thesocket"})
            t("socket:" + socket_dir + "/thesocket", {"display_name": "socket:" + socket_dir + "/thesocket"})
        e("tcp://host:NOTANUMBER/")
        e("tcp://host:-1/")
        e("tcp://host:65536/")
        t("tcp://username@host/", {"username": "username", "password": None})
        for socktype in ("tcp", "ws", "wss", "ssl", "ssh"):
            # e(socktype+"://a/b/c/d")
            t(socktype + "://username:password@host:10000/DISPLAY?key1=value1", {
                "type": socktype,
                "display": "DISPLAY",
                "key1": "value1",
                "username": "username",
                "password": "password",
                "port": 10000,
                "host": "host",
                "local": False,
            })
        t("tcp://fe80::c1:ac45:7351:ea69:14500", {"host": "fe80::c1:ac45:7351:ea69", "port": 14500})
        t("tcp://fe80::c1:ac45:7351:ea69%eth1:14500", {"host": "fe80::c1:ac45:7351:ea69%eth1", "port": 14500})
        t("tcp://[fe80::c1:ac45:7351:ea69]:14500", {"host": "fe80::c1:ac45:7351:ea69", "port": 14500})
        t("tcp://host/100,key1=value1", {"key1": "value1"})
        t("tcp://host/key1=value1", {"key1": "value1"})
        try:
            # pylint: disable=import-outside-toplevel
            from xpra.net.vsock import CID_ANY, PORT_ANY
        except ImportError:
            pass
        else:
            t("vsock://any:any/", {"vsock": (CID_ANY, PORT_ANY)})
            t("vsock://10:2000/", {"vsock": (10, 2000)})
        t("vnc+ssh://host/0")
        with IgnoreWarningsContext():
            t("tcp:localhost:10000", {"host": "localhost", "port": 10000, "type": "tcp"})


def main():
    unittest.main()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from ctypes import (
    WinError, get_last_error,
    byref, c_void_p, c_char, sizeof, memmove,
)

from xpra.platform.win32 import constants as win32con
from xpra.platform.win32.common import (
    ICONINFO, BITMAPV5HEADER, HICON, HBITMAP,
    CreateIconIndirect,
    GetDC, ReleaseDC,
    CreateBitmap, CreateDIBSection,
    DeleteObject,
)
from xpra.log import Logger

log = Logger("win32")

BI_RGB = 0
BI_BITFIELDS = 0x00000003


def image_to_ICONINFO(img, alpha=True) -> HICON:
    w, h = img.size
    if alpha and img.mode.find("A") >= 0:  # ie: RGBA
        rgb_format = "BGRA"
    else:
        rgb_format = "BGR"
    rgb_data = img.tobytes("raw", rgb_format)
    return make_ICONINFO(w, h, rgb_data, rgb_format=rgb_format)


def make_ICONINFO(w: int, h: int, rgb_data: bytes, rgb_format="BGRA") -> HICON:
    log("make_ICONINFO(%i, %i, %i bytes, %s)", w, h, len(rgb_data), rgb_format)
    bitmap: HBITMAP = HBITMAP(0)
    mask: HBITMAP = HBITMAP(0)
    try:
        bytes_per_pixel = len(rgb_format)
        bitmap = rgb_to_bitmap(rgb_data, bytes_per_pixel, w, h)
        log("rgb_to_bitmap(%i bytes, %i, %i, %i)=%s", len(rgb_data), bytes_per_pixel, w, h, bitmap)
        mask = CreateBitmap(w, h, 1, 1, None)
        log("CreateBitmap(%i, %i, 1, 1, None)=%#x", w, h, mask or 0)
        if not mask:
            raise WinError(get_last_error())
        iconinfo = ICONINFO()
        iconinfo.fIcon = True
        iconinfo.hbmMask = mask
        iconinfo.hbmColor = bitmap
        hicon = CreateIconIndirect(byref(iconinfo))
        log("CreateIconIndirect()=%#x", hicon or 0)
        if not hicon:
            raise WinError(get_last_error())
        return hicon
    except Exception:
        log.error("Error: failed to set tray icon", exc_info=True)
        return HICON(0)
    finally:
        if mask:
            DeleteObject(mask)
        if bitmap:
            DeleteObject(bitmap)


def rgb_to_bitmap(rgb_data: bytes, bytes_per_pixel: int, w: int, h: int) -> HBITMAP:
    log("rgb_to_bitmap%s", (rgb_data, bytes_per_pixel, w, h))
    assert bytes_per_pixel in (3, 4)  # only BGRA or BGR are supported
    assert w > 0 and h > 0
    header = BITMAPV5HEADER()
    header.bV5Size = sizeof(BITMAPV5HEADER)
    header.bV5Width = w
    header.bV5Height = -h
    header.bV5Planes = 1
    header.bV5BitCount = bytes_per_pixel * 8
    header.bV5Compression = BI_RGB  # BI_BITFIELDS
    # header.bV5RedMask = 0x000000ff
    # header.bV5GreenMask = 0x0000ff00
    # header.bV5BlueMask = 0x00ff0000
    # header.bV5AlphaMask = 0xff000000
    hdc = 0
    try:
        hdc = GetDC(None)
        dataptr = c_void_p()
        log("GetDC()=%#x", hdc)
        bitmap = CreateDIBSection(hdc, byref(header), win32con.DIB_RGB_COLORS, byref(dataptr), None, 0)
    finally:
        if hdc:
            ReleaseDC(None, hdc)
    if not dataptr or not bitmap:
        raise WinError(get_last_error())
    log("CreateDIBSection(..) got bitmap=%#x, dataptr=%s", int(bitmap), dataptr)
    size = len(rgb_data)
    # noinspection PyTypeChecker
    buftype = c_char * size
    # noinspection PyCallingNonCallable
    buf = buftype()
    buf.value = rgb_data
    memmove(dataptr, byref(buf), w * h * bytes_per_pixel)
    return bitmap



# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from collections.abc import Sequence

# The data for this table can be found mostly here:
# http://msdn.microsoft.com/en-us/library/aa912040.aspx
# and here:
# http://support.microsoft.com/kb/278957
# Format:
# Language identifier: (Language code, Sublanguage - locale, Language, Default code page, X11 keymap, x11 variants)
# The x11 keymap name was found in /usr/share/X11/xkb/rules/*
# This is used for converting the layout we detect using win32api into
# something that can be used by X11 (a layout with optional variant)
UNICODE = -1
LATAM_VARIANTS: Sequence[str] = ("nodeadkeys", "deadtilde", "sundeadkeys")
ARA_VARIANTS: Sequence[str] = ("azerty", "azerty_digits", "digits", "qwerty", "qwerty_digits", "buckwalter")
ES_VARIANTS: Sequence[str] = ("nodeadkeys", "deadtilde", "sundeadkeys", "dvorak", "est", "cat", "mac")
RS_VARIANTS: Sequence[str] = ("yz", "latin", "latinunicode", "latinyz", "latinunicodeyz", "alternatequotes",
                              "latinalternatequotes", "rue")
FR_VARIANTS: Sequence[str] = ("nodeadkeys", "sundeadkeys", "oss", "oss_latin9", "oss_nodeadkeys", "oss_sundeadkeys",
                              "latin9", "latin9_nodeadkeys", "latin9_sundeadkeys", "bepo", "bepo_latin9", "dvorak",
                              "mac", "bre", "oci", "geo")
WIN32_LAYOUTS: dict[int, tuple[str, str, str, int, str, Sequence[str]]] = {
    1025: ("ARA", "Saudi Arabia",   "Arabic",                   1356,   "ar", ()),
    1026: ("BGR", "Bulgaria",       "Bulgarian",                1251,   "bg", ("phonetic", "bas_phonetic")),
    1027: ("CAT", "Spain",          "Catalan",                  1252,   "ad", ()),
    1028: ("CHT", "Taiwan",         "Chinese",                  950,    "tw", ("indigenous", "saisiyat")),
    1029: ("CSY", "Czech",          "Czech",                    1250,   "cz", ("bksl", "qwerty", "qwerty_bksl", "ucw", "dvorak-ucw")),
    1030: ("DAN", "Denmark",        "Danish",                   1252,   "dk", ("nodeadkeys", "mac", "mac_nodeadkeys", "dvorak")),
    1031: ("DEU", "Germany",        "German",                   1252,   "de", ("nodeadkeys", "sundeadkeys", "mac")),
    1032: ("ELL", "Greece",         "Greek",                    1253,   "gr", ("simple", "extended", "nodeadkeys", "polytonic")),
    1033: ("USA", "United States",  "English",                  1252,   "us", ()),
    1034: ("ESP", "Spain (Traditional sort)", "Spanish",        1252,   "es", ES_VARIANTS),
    1035: ("FIN", "Finland",        "Finnish",                  1252,   "fi", ("classic", "nodeadkeys", "smi", "mac")),
    1036: ("FRA", "France",         "French",                   1252,   "fr", FR_VARIANTS),
    1037: ("HEB", "Israel",         "Hebrew",                   1255,   "il", ("lyx", "phonetic", "biblical")),
    1038: ("HUN", "Hungary",        "Hungarian",                1250,   "hu", ("standard", "nodeadkeys", "qwerty", "101_qwertz_comma_dead", "101_qwertz_comma_nodead", "101_qwertz_dot_dead", "101_qwertz_dot_nodead", "101_qwerty_comma_dead", "101_qwerty_comma_nodead", "101_qwerty_dot_dead", "101_qwerty_dot_nodead", "102_qwertz_comma_dead", "102_qwertz_comma_nodead", "102_qwertz_dot_dead", "102_qwertz_dot_nodead", "102_qwerty_comma_dead", "102_qwerty_comma_nodead", "102_qwerty_dot_dead", "102_qwerty_dot_nodead")),
    1039: ("ISL", "Iceland",        "Icelandic",                1252,   "is", ("sundeadkeys", "nodeadkeys", "mac", "dvorak")),
    1040: ("ITA", "Italy",          "Italian",                  1252,   "it", ("nodeadkeys", "mac", "us", "geo")),
    1041: ("JPN", "Japan",          "Japanese",                 932,    "jp", ("kana", "kana86", "OADG109A", "mac")),
    1042: ("KOR", "Korea",          "Korean",                   949,    "kr", ("kr104", )),
    1043: ("NLD", "Netherlands",    "Dutch",                    1252,   "nl", ("sundeadkeys", "mac", "std")),
    1044: ("NOR", "Norway (Bokmål)","Norwegian",                1252,   "no", ("nodeadkeys", "dvorak", "smi", "smi_nodeadkeys", "mac", "mac_nodeadkeys")),
    1045: ("PLK", "Poland",         "Polish",                   1250,   "pl", ("qwertz", "dvorak", "dvorak_quotes", "dvorak_altquotes", "csb", "ru_phonetic_dvorak", "dvp")),
    1046: ("PTB", "Brazil",         "Portuguese",               1252,   "br", ("nodeadkeys", "dvorak", "nativo", "nativo-us", "nativo-epo")),
    1048: ("ROM", "Romania",        "Romanian",                 1250,   "ro", ("cedilla", "std", "std_cedilla", "winkeys")),
    1049: ("RUS", "Russia",         "Russian",                  1251,   "ru", ("phonetic", "phonetic_winkeys", "typewriter", "legacy", "typewriter-legacy", "tt", "os_legacy", "os_winkeys", "cv", "cv_latin", "udm", "kom", "sah", "xal", "dos", "srp", "bak", "chm")),
    1050: ("HRV", "Croatia",        "Croatian",                 1250,   "hr", ("alternatequotes", "unicode", "unicodeus", "us")),
    1051: ("SKY", "Slovakia",       "Slovakian",                1250,   "sk", ("bksl", "qwerty", "qwerty_bksl")),
    1052: ("SQI", "Albania",        "Albanian",                 1250,   "al", ()),
    1053: ("SVE", "Sweden",         "Swedish",                  1252,   "se", ("nodeadkeys", "dvorak", "rus", "rus_nodeadkeys", "smi", "mac", "svdvorak", "swl")),
    1054: ("THA", "Thailand",       "Thai",                     874,    "th", ("tis", "pat")),
    1055: ("TRK", "Turkey",         "Turkish",                  1254,   "tr", ("f", "alt", "sundeadkeys", "ku", "ku_f", "ku_alt", "intl", "crh", "crh_f", "crh_alt")),
    1056: ("URP", "Pakistan",       "Urdu",                     1256,   "pk", ("urd-crulp", "urd-nla", "ara", "snd")),
    1057: ("IND", "Indonesia (Bahasa)", "Indonesian",           1252,   "",   (), ),
    1058: ("UKR", "Ukraine",        "Ukrainian",                1251,   "ua", ("phonetic", "typewriter", "winkeys", "legacy", "rstu", "rstu_ru", "homophonic")),
    1059: ("BEL", "Belarus",        "Belarusian",               1251,   "by", ("legacy", "latin")),
    1060: ("SLV", "Slovenia",       "Slovenian",                1250,   "si", ("alternatequotes", "us")),
    1061: ("ETI", "Estonia",        "Estonian",                 1257,   "ee", ("nodeadkeys", "dvorak", "us")),
    1062: ("LVI", "Latvia",         "Latvian",                  1257,   "lv", ("apostrophe", "tilde", "fkey", "modern", "ergonomic", "adapted")),
    1063: ("LTH", "Lithuania",      "Lithuanian",               1257,   "lt", ("std", "us", "ibm", "lekp", "lekpa")),
    1065: ("FAR", "Iran",           "Farsi",                    1256,   "ir", ()),
    1066: ("VIT", "Viet Nam",       "Vietnamese",               1258,   "vn", ()),
    1067: ("HYE", "Armenia",        "Armenian",                 UNICODE,"am", ("phonetic", "phonetic-alt", "eastern", "western", "eastern-alt")),
    1068: ("AZE", "Azerbaijan (Latin)", "Azeri",                1254,   "az", ("cyrillic", )),
    1069: ("EUQ", "Spain",          "Basque",                   1252,   "es", ()),
    1071: ("MKI", "F.Y.R.O. Macedonia", "F.Y.R.O. Macedonia",   1251,   "mk", ("nodeadkeys", )),
    1078: ("AFK", "South Africa",   "Afrikaans",                1252,   "",   ()),
    1079: ("KAT", "Georgia",        "Georgian",                 UNICODE,"ge", ("ergonomic", "mess", "ru", "os")),
    1080: ("FOS", "Faroe Islands",  "Faroese",                  1252,   "fo", ("nodeadkeys", )),
    1081: ("HIN", "India",          "Hindi",                    UNICODE,"in", ("bolnagri", "hin-wx")),
    1086: ("MSL", "Malaysia",       "Malay",                    1252,   "in", ("mal", "mal_lalitha", "mal_enhanced")),
    1087: ("KKZ", "Kazakhstan",      "Kazakh",                   1251,   "kz", ("ruskaz", "kazrus")),
    1088: ("KYR", "Kyrgyzstan",     "Kyrgyz",                   1251,   "kg", ("phonetic", )),
    1089: ("SWK", "Kenya",          "Swahili",                  1252,   "ke", ("kik", )),
    1091: ("UZB", "Uzbekistan (Latin)", "Uzbek",                1254,   "uz", ("latin", )),
    1092: ("TTT", "Tatarstan",      "Tatar",                    1251,   "ru", ("tt", )),
    1094: ("PAN", "India (Gurmukhi script)", "Punjabi",         UNICODE,"in", ("guru", "jhelum")),
    1095: ("GUJ", "India",          "Gujarati",                 UNICODE,"in", ("guj", )),
    1097: ("TAM", "India",          "Tamil",                    UNICODE,"in", ("tam_unicode", "tam_keyboard_with_numerals", "tam_TAB", "tam_TSCII", "tam")),
    1098: ("TEL", "India (Telugu script)", "Telugu",            UNICODE,"in", ("tel", )),
    1099: ("KAN", "India (Kannada script)", "Kannada",          UNICODE,"in", ("kan", )),
    1102: ("MAR", "India",          "Marathi",                  UNICODE,"in", ()),
    1103: ("SAN", "India",          "Sanskrit",                 UNICODE,"in", ()),
    1104: ("MON", "Mongolia",       "Mongolian (Cyrillic)",     1251,   "mn", ()),
    1110: ("GLC", "Spain",          "Galician",                 1252,   "es", ()),
    1111: ("KNK", "India",          "Konkani",                  UNICODE,"in", ()),
    1114: ("SYR", "Syria",          "Syriac",                   UNICODE,"sy", ("syc", "syc_phonetic", "ku", "ku_f", "ku_alt")),
    1125: ("DIV", "Maldives",       "Divehi",                   UNICODE,"",   ()),
    2049: ("ARI", "Iraq",           "Arabic",                   1256,   "iq", ("ku", "ku_f", "ku_alt", "ku_ara")),
    2052: ("CHS", "PRC",            "Chinese, Simplified",      0,      "cn", ("tib", "tib_asciinum", "uig")),
    2055: ("DES", "Switzerland",    "German",                   1252,   "de", ("deadacute", "deadgraveacute", "nodeadkeys", "ro", "ro_nodeadkeys", "dvorak", "sundeadkeys", "neo", "mac", "mac_nodeadkeys", "dsb", "dsb_qwertz", "qwerty", "ru")),
    2057: ("ENG", "UK",             "English",                  1252,   "gb", ("extd", "intl", "dvorak", "dvorakukp", "mac", "mac_intl", "colemak")),
    2058: ("ESM", "Mexico",         "Spanish",                  1252,   "es", ES_VARIANTS),
    2060: ("FRB", "Benelux",        "French",                   1252,   "be", ("oss", "oss_latin9", "oss_sundeadkeys", "iso-alternate", "nodeadkeys", "sundeadkeys", "wang")),
    2064: ("ITS", "Switzerland",    "Italian",                  1252,   "it", ("nodeadkeys", "mac", "us", "geo")),
    2067: ("NLB", "Belgium",        "Dutch",                    1252,   "nl", ("sundeadkeys", "mac", "std")),
    2068: ("NON", "Norway (Nynorsk)", "Norwegian",              1252,   "no", ("nodeadkeys", "dvorak", "smi", "smi_nodeadkeys", "mac", "mac_nodeadkeys")),
    2070: ("PTG", "Portugal",       "Portuguese",               1252,   "pt", ("nodeadkeys", "sundeadkeys", "mac", "mac_nodeadkeys", "mac_sundeadkeys", "nativo", "nativo-us", "nativo-epo")),
    2074: ("SRL", "Serbia (Latin)", "Serbian",                  1250,   "rs", RS_VARIANTS),
    2077: ("SVF", "Finland",        "Swedish",                  1252,   "se", ("nodeadkeys", "dvorak", "rus", "rus_nodeadkeys", "smi", "mac", "svdvorak", "swl")),
    2092: ("AZE", "Azerbaijan (Cyrillic)", "Azeri",             1251,   "az", ("cyrillic", )),
    2110: ("MSB", "Brunei Darussalam", "Malay",                 1252,   "in", ("mal", "mal_lalitha", "mal_enhanced")),
    2115: ("UZB", "Uzbekistan (Cyrillic)", "Uzbek",             1251,   "uz", ("latin", )),
    3073: ("ARE", "Egypt",          "Arabic",                   1256,   "ara", ARA_VARIANTS),
    3076: ("ZHH", "Hong Kong SAR",  "Chinese",                  950,    "cn", ()),
    3079: ("DEA", "Austria",        "German",                   1252,   "at", ("nodeadkeys", "sundeadkeys", "mac")),
    3081: ("ENA", "Australia",      "English",                  1252,   "us", ()),
    3082: ("ESN", "Spain (International sort)", "Spanish",      1252,   "es", ES_VARIANTS),
    3084: ("FRC", "Canada",         "French",                   1252,   "ca", ("fr-dvorak", "fr-legacy", "multix", "multi", "multi-2gr", "ike")),
    3098: ("SRB", "Serbia (Cyrillic)", "Serbian",               1251,   "", RS_VARIANTS),
    4097: ("ARL", "Libya",          "Arabic",                   1256,   "ara", ARA_VARIANTS),
    4100: ("ZHI", "Singapore",      "Chinese",                  936,    "cn", ()),
    4103: ("DEL", "Luxembourg",     "German",                   1252,   "de", ()),
    4105: ("ENC", "Canada",         "English",                  1252,   "ca", ("eng", )),
    4106: ("ESG", "Guatemala",      "Spanish",                  1252,   "latam", LATAM_VARIANTS),
    4108: ("FRS", "Switzerland",    "French",                   1252,   "ch", ("fr", "fr_nodeadkeys", "fr_sundeadkeys", "fr_mac")),
    5121: ("ARG", "Algeria",        "Arabic",                   1256,   "ara", ARA_VARIANTS),
    5124: ("ZHM", "Macao SAR",      "Chinese",                  950,    "cn", ()),
    5127: ("DEC", "Liechtenstein",  "German",                   1252,   "de", ()),
    5129: ("ENZ", "New Zealand",    "English",                  1252,   "us", ()),
    5130: ("ESC", "Costa Rica",     "Spanish",                  1252,   "latam", LATAM_VARIANTS),
    5132: ("FRL", "Luxembourg",     "French",                   1252,   "fr", FR_VARIANTS),
    6145: ("ARM", "Morocco",        "Arabic",                   1256,   "ara", ARA_VARIANTS),
    6153: ("ENI", "Ireland",        "English",                  1252,   "en", ()),
    6154: ("ESA", "Panama",         "Spanish",                  1252,   "latam", LATAM_VARIANTS),
    6156: ("FRM", "Monaco",         "French",                   1252,   "fr", FR_VARIANTS),
    7169: ("ART", "Tunisia",        "Arabic",                   1256,   "ara", ARA_VARIANTS),
    7177: ("ENS", "South Africa",   "English",                  1252,   "en", ()),
    7178: ("ESD", "Dominican Republic", "Spanish",              1252,   "latam", LATAM_VARIANTS),
    8193: ("ARO", "Oman",           "Arabic",                   1256,   "ara", ARA_VARIANTS),
    8201: ("ENJ", "Jamaica",        "English",                  1252,   "en", ()),
    8202: ("ESV", "Venezuela",      "Spanish",                  1252,   "latam", LATAM_VARIANTS),
    9217: ("ARY", "Yemen",          "Arabic",                   1256,   "ara", ARA_VARIANTS),
    9225: ("ENB", "Caribbean",      "English",                  1252,   "en", ()),
    9226: ("ESO", "Colombia",       "Spanish",                  1252,   "latam", LATAM_VARIANTS),
    10241: ("ARS", "Syria",         "Arabic",                   1256,   "sy", ("syc", "syc_phonetic")),
    10249: ("ENL", "Belize",        "English",                  1252,   "us", ()),
    10250: ("ESR", "Peru",          "Spanish",                  1252,   "latam", LATAM_VARIANTS),
    11265: ("ARJ", "Jordan",        "Arabic",                   1256,   "ara", ARA_VARIANTS),
    11273: ("ENT", "Trinidad",      "English",                  1252,   "us", ()),
    11274: ("ESS", "Argentina",     "Spanish",                  1252,   "latam", LATAM_VARIANTS),
    12289: ("ARB", "Lebanon",       "Arabic",                   1256,   "ara", ARA_VARIANTS),
    12297: ("ENW", "Zimbabwe",      "English",                  1252,   "us", ()),
    12298: ("ESF", "Ecuador",       "Spanish",                  1252,   "latam", LATAM_VARIANTS),
    13321: ("ENP", "Philippines",   "English",                  1252,   "us", ()),
    13313: ("ARK", "Kuwait",        "Arabic",                   1256,   "ara", ARA_VARIANTS),
    13322: ("ESL", "Chile",         "Spanish",                  1252,   "latam", LATAM_VARIANTS),
    14337: ("ARU", "U.A.E.",        "Arabic",                   1256,   "ara", ARA_VARIANTS),
    14345: ("",    "Indonesia",     "English",                  1252,   "us", ()),
    14346: ("ESY", "Uruguay",       "Spanish",                  1252,   "latam", LATAM_VARIANTS),
    15361: ("ARH", "Bahrain",       "Arabic",                   1256,   "ara", ARA_VARIANTS),
    15369: ("ZHH", "Hong Kong SAR", "English",                  1252,   "us", ()),
    15370: ("ESZ", "Paraguay",      "Spanish",                  1252,   "latam", LATAM_VARIANTS),
    16385: ("ARQ", "Qatar",         "Arabic",                   1256,   "ara", ARA_VARIANTS),
    16393: ("",    "India",         "English",                  1252,   "us", ()),
    16394: ("ESB", "Bolivia",       "Spanish",                  1252,   "latam", LATAM_VARIANTS),
    17417: ("",    "Malaysia",      "English",                  1252,   "us", ()),
    17418: ("ESE", "El Salvador",   "Spanish",                  1252,   "latam", LATAM_VARIANTS),
    18441: ("",    "Singapore",     "English",                  1252,   "us", ()),
    18442: ("ESH", "Honduras",      "Spanish",                  1252,   "latam", LATAM_VARIANTS),
    19466: ("ESI", "Nicaragua",     "Spanish",                  1252,   "latam", LATAM_VARIANTS),
    20490: ("ESU", "Puerto Rico",   "Spanish",                  1252,   "latam", LATAM_VARIANTS),
    58378: ("",    "LatAm",         "Spanish",                  1252,   "latam", LATAM_VARIANTS),
    58380: ("",    "North Africa",  "French",                   1252,   "fr", FR_VARIANTS),
}

# map win32 keyboard codes to x11 names:
# based on
# https://docs.microsoft.com/en-us/windows-hardware/manufacture/desktop/windows-language-pack-default-values
WIN32_KEYBOARDS: dict[int, tuple[str, str]] = {
    0x0000041c: ("al", "Albania"),
    0x00000401: ("ar", "Arabic (101)"),
    0x00010401: ("ar", "Arabic (102)"),
    0x00020401: ("ar", "Arabic (102) AZERTY"),
    0x0000042b: ("am", "Armenian Eastern"),
    0x0002042b: ("am", "Armenian Phonetic"),
    0x0003042b: ("am", "Armenian Typewriter"),
    0x0001042b: ("am", "Armenian Western"),
    0x0000044d: ("in", "Assamese - Inscript"),
    0x0001042c: ("az", "Azerbaijani (Standard)"),
    0x0000082c: ("az", "Azerbaijani Cyrillic"),
    0x0000042c: ("az", "Azerbaijani Latin"),
    0x0000046d: ("ru", "Bashkir"),
    0x00000423: ("by", "Belarusian"),
    0x0001080c: ("be", "Belgian (Comma)"),
    0x00000813: ("be", "Belgian (Period)"),
    0x0000080c: ("be", "Belgian French"),
    0x00000445: ("bd", "Bangla (Bangladesh)"),
    0x00020445: ("bd", "Bangla (India)"),
    0x00010445: ("bd", "Bangla (India - Legacy)"),
    0x0000201a: ("ba", "Bosnian (Cyrillic)"),
    0x000b0c00: ("??", "Buginese"),
    0x00030402: ("bg", "Bulgarian"),
    0x00010402: ("bg", "Bulgarian (Latin)"),
    0x00020402: ("bg", "Bulgarian (phonetic layout)"),
    0x00040402: ("bg", "Bulgarian (phonetic traditional)"),
    0x00000402: ("bg", "Bulgarian (Typewriter)"),
    0x00001009: ("ca", "Canadian French"),
    0x00000c0c: ("ca", "Canadian French (Legacy)"),
    0x00011009: ("ca", "Canadian Multilingual Standard"),
    0x0000085f: ("fr", "Central Atlas Tamazight"),
    0x00000492: ("ku", "Central Kurdish"),
    0x0000045c: ("us", "Cherokee Nation"),
    0x0001045c: ("us", "Cherokee Nation Phonetic"),
    0x00000804: ("cn", "Chinese (Simplified) - US Keyboard"),
    0x00000404: ("cn", "Chinese (Traditional) - US Keyboard"),
    0x00000c04: ("cn", "Chinese (Traditional, Hong Kong S.A.R.)"),
    0x00001404: ("cn", "Chinese (Traditional Macao S.A.R.) US Keyboard"),
    0x00001004: ("cn", "Chinese (Simplified, Singapore) - US keyboard"),
    0x0000041a: ("hr", "Croatian"),
    0x00000405: ("cz", "Czech"),
    0x00010405: ("cz", "Czech (QWERTY)"),
    0x00020405: ("cz", "Czech Programmers"),
    0x00000406: ("dk", "Danish"),
    0x00000439: ("in", "Devanagari-INSCRIPT"),
    0x00000465: ("in", "Divehi Phonetic"),
    0x00010465: ("in", "Divehi Typewriter"),
    0x00000413: ("nl", "Dutch"),
    0x00000C51: ("dz", "Dzongkha"),
    0x00000425: ("ee", "Estonian"),
    0x00000438: ("fo", "Faeroese"),
    0x0000040b: ("fi", "Finnish"),
    0x0001083b: ("fi", "Finnish with Sami"),
    0x0000040c: ("fr", "French"),
    0x00120c00: ("??", "Futhark"),
    0x00000437: ("ge", "Georgian"),
    0x00020437: ("ge", "Georgian (Ergonomic)"),
    0x00010437: ("ge", "Georgian (QWERTY)"),
    0x00030437: ("ge", "Georgian Ministry of Education and Science Schools"),
    0x00040437: ("ge", "Georgian (Old Alphabets)"),
    0x00000407: ("de", "German"),
    0x00010407: ("de", "German (IBM)"),
    0x000c0c00: ("??", "Gothic"),
    0x00000408: ("gr", "Greek"),
    # Duplicate of 'tr':
    # 0x00010408: ("gr", "Greek (220)"),
    0x00030408: ("gr", "Greek (220) Latin"),
    0x00020408: ("gr", "Greek (319)"),
    0x00040408: ("gr", "Greek (319) Latin"),
    0x00050408: ("gr", "Greek Latin"),
    0x00060408: ("gr", "Greek Polytonic"),
    0x0000046f: ("??", "Greenlandic"),
    0x00000474: ("??", "Guarani"),
    0x00000447: ("in", "Gujarati"),
    0x00000468: ("gh", "Hausa"),
    0x0000040d: ("il", "Hebrew"),
    0x00010439: ("in", "Hindi Traditional"),
    0x0000040e: ("hu", "Hungarian"),
    0x0001040e: ("hu", "Hungarian 101-key"),
    0x0000040f: ("is", "Icelandic"),
    0x00000470: ("ng", "Igbo"),
    0x00004009: ("in", "India"),
    0x0000085d: ("ca", "Inuktitut - Latin"),
    0x0001045d: ("ca", "Inuktitut - Naqittaut"),
    0x00001809: ("ie", "Ireland"),
    0x00000410: ("it", "Italian"),
    0x00010410: ("it", "Italian"),
    0x00000411: ("jp", "Japanese"),
    0x00110c00: ("??", "Javanese"),
    0x0000044b: ("in", "Kannada"),
    0x0000043f: ("kz", "Kazakh"),
    0x00000453: ("kh", "Khmer"),
    0x00010453: ("kh", "Khmer (NIDA)"),
    0x00000412: ("kr", "Korean"),
    0x00000440: ("kg", "Kyrgyz Cyrillic"),
    0x00000454: ("la", "Lao"),
    0x0000080a: ("latam", "Latin American"),
    0x00020426: ("lv", "Latvian (Standard)"),
    0x00010426: ("lv", "Latvian (Legacy)"),
    0x00070c00: ("??", "Lisu (Basic)"),
    0x00080c00: ("??", "Lisu (Standard)"),
    0x00010427: ("lt", "Lithuanian"),
    0x00000427: ("lt", "Lithuanian IBM"),
    0x00020427: ("lt", "Lithuanian Standard"),
    0x0000046e: ("de", "Luxembourgish"),
    0x0000042f: ("mk", "Macedonia (FYROM)"),
    0x0001042f: ("mk", "Macedonia (FYROM) - Standard"),
    0x0000044c: ("in", "Malayalam"),
    0x0000043a: ("mt", "Maltese 47-Key"),
    0x0001043a: ("mt", "Maltese 48-key"),
    0x00000481: ("mao", "Maori"),
    0x0000044e: ("in", "Marathi"),
    0x00000850: ("mn", "Mongolian (Mongolian Script - Legacy)"),
    0x00020850: ("mn", "Mongolian (Mongolian Script - Standard)"),
    0x00000450: ("mn", "Mongolian Cyrillic"),
    0x00010c00: ("mm", "Myanmar"),
    0x00090c00: ("??", "N'ko"),
    0x00000461: ("np", "Nepali"),
    0x00020c00: ("th", "New Tai Lue"),
    0x00000414: ("no", "Norwegian"),
    0x0000043b: ("no", "Norwegian with Sami"),
    0x00000448: ("in", "Odia"),
    0x000d0c00: ("in", "Ol Chiki"),
    0x000f0c00: ("??", "Old Italic"),
    0x000e0c00: ("??", "Osmanya"),
    0x00000463: ("af", "Pashto (Afghanistan)"),
    0x00000429: ("ir", "Persian"),
    0x00050429: ("ir", "Persian (Standard)"),
    0x000a0c00: ("cn", "Phags-pa"),
    0x00010415: ("pl", "Polish (214)"),
    0x00000415: ("pl", "Polish (Programmers)"),
    0x00000816: ("pt", "Portuguese"),
    0x00000416: ("pt", "Portuguese (Brazilian ABNT),"),
    0x00010416: ("pt", "Portuguese (Brazilian ABNT2)"),
    0x00000446: ("in", "Punjabi"),
    0x00000418: ("ro", "Romanian (Legacy)"),
    0x00020418: ("ro", "Romanian (Programmers)"),
    0x00010418: ("ro", "Romanian (Standard)"),
    0x00000419: ("ru", "Russian"),
    0x00020419: ("ru", "Russian - Mnemonic"),
    0x00010419: ("ru", "Russian (Typewriter)"),
    # duplicate of 'tr'
    # 0x00000485: ("ru", "Sakha"),
    0x0002083b: ("fi", "Sami Extended Finland-Sweden"),
    0x0001043b: ("no", "Sami Extended Norway"),
    0x00011809: ("gb", "Scottish Gaelic"),
    0x00000c1a: ("ru", "Serbian (Cyrillic)"),
    0x0000081a: ("rs", "Serbian (Latin)"),
    0x0000046c: ("za", "Sesotho sa Leboa"),
    0x00000432: ("za", "Setswana"),
    0x0000045b: ("in", "Sinhala"),
    0x0001045b: ("in", "Sinhala - wij 9"),
    0x0000041b: ("sk", "Slovak"),
    0x0001041b: ("sk", "Slovak (QWERTY)"),
    0x00000424: ("si", "Slovenian"),
    0x00100c00: ("in", "Sora"),
    0x0001042e: ("de", "Sorbian Extended"),
    0x0002042e: ("de", "Sorbian Standard"),
    0x0000042e: ("de", "Sorbian Standard (Legacy)"),
    0x0000040a: ("es", "Spanish"),
    0x0001040a: ("es", "Spanish Variation"),
    0x0000041d: ("se", "Swedish"),
    0x0000083b: ("se", "Swedish with Sami"),
    0x0000100c: ("fr", "Swiss French"),
    0x00000807: ("de", "Swiss German"),
    0x0000045a: ("sy", "Syriac"),
    0x0001045a: ("sy", "Syriac Phonetic"),
    0x00030c00: ("cn", "Tai Le"),
    0x00000428: ("tj", "Tajik"),
    0x00000449: ("in", "Tamil"),
    0x00010444: ("ru", "Tatar"),
    0x00000444: ("ru", "Tatar (Legacy)"),
    0x0000044a: ("in", "Telugu"),
    0x0000041e: ("th", "Thai Kedmanee"),
    0x0002041e: ("th", "Thai Kedmanee (non-ShiftLock)"),
    0x0001041e: ("th", "Thai Pattachote"),
    0x0003041e: ("th", "Thai Pattachote (non-ShiftLock)"),
    0x00010451: ("cn", "Tibetan (PRC - Standard)"),
    0x00000451: ("cn", "Tibetan (PRC - Legacy)"),
    0x00050c00: ("ma", "Tifinagh (Basic)"),
    0x00060c00: ("ma", "Tifinagh (Full)"),
    0x0001041f: ("tr", "Turkish F"),
    0x0000041f: ("tr", "Turkish Q"),
    0x00000442: ("tm", "Turkmen"),
    0x00010408: ("tr", "Uyghur    "),
    0x00000480: ("tr", "Uyghur (Legacy)"),
    0x00000422: ("ua", "Ukrainian"),
    0x00020422: ("ua", "Ukrainian (Enhanced)"),
    0x00000809: ("gb", "United Kingdom"),
    0x00000452: ("gb", "United Kingdom Extended"),
    0x00010409: ("us", "United States - Dvorak"),
    0x00020409: ("us", "United States - International"),
    0x00030409: ("us", "United States-Dvorak for left hand"),
    0x00040409: ("us", "United States-Dvorak for right hand"),
    0x00000409: ("us", "United States - English"),
    0x00000420: ("in", "Urdu"),
    0x00000843: ("ru", "Uzbek Cyrillic"),
    0x0000042a: ("vn", "Vietnamese"),
    0x00000488: ("fr", "Wolof"),
    0x00000485: ("tr", "Yakut"),
    0x0000046a: ("??", "Yoruba"),
}

# This is generated from the table above,
# so we can let the user choose his own layout.
# (country,language) : (layout,variant)
X11_LAYOUTS: dict[tuple[str, str], tuple[str, Sequence[str]]] = {}
for ccode, country, language, codepage, layout, variants in WIN32_LAYOUTS.values():
    X11_LAYOUTS[(country, language)] = (layout, variants)
LAYOUT_VARIANTS: dict[str, list[str]] = {}
for win32layout in WIN32_LAYOUTS.values():
    layout, variants = win32layout[-2:]
    if not layout:
        continue
    layout_variants = LAYOUT_VARIANTS.setdefault(layout, [])
    for variant in variants:
        if variant not in layout_variants:
            layout_variants.append(variant)


# make it easier to resolve a 2-letter code into a meaningful description:
LAYOUT_NAMES: dict[str, str] = {}
for code3, country, name, codepage, code2, variants in WIN32_LAYOUTS.values():
    if code2 in LAYOUT_NAMES:
        continue
    LAYOUT_NAMES[code2] = name


def parse_xkbmap_query(xkbmap_query: str) -> dict[str, str]:
    """ parses the output of "setxkbmap -query" into a dict """
    import re
    settings = {}
    opt_re = re.compile(r"(\w*):\s*(.*)")
    for line in xkbmap_query.splitlines():
        m = opt_re.match(line)
        if m:
            v = m.group(2).strip()
            if v.strip(","):
                settings[m.group(1)] = v
    return settings



# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
import glob
import posixpath
import mimetypes
import socket
from urllib.parse import unquote
from http.server import BaseHTTPRequestHandler
from typing import Any
from collections.abc import Iterable, Callable

from xpra.common import FULL_INFO
from xpra.net.common import HttpResponse
from xpra.net.http.common import EXTENSION_TO_MIMETYPE
from xpra.net.http.directory_listing import list_directory
from xpra.net.bytestreams import pretty_socket
from xpra.util.objects import AdHocStruct
from xpra.util.str_fn import std, csv, repr_ellipsized
from xpra.util.env import envbool
from xpra.platform.paths import get_desktop_background_paths
from xpra.log import Logger

log = Logger("http")

HTTP_ACCEPT_ENCODING = os.environ.get("XPRA_HTTP_ACCEPT_ENCODING", "br,gzip").split(",")
DIRECTORY_LISTING = envbool("XPRA_HTTP_DIRECTORY_LISTING", False)

AUTH_REALM = os.environ.get("XPRA_HTTP_AUTH_REALM", "Xpra")
AUTH_USERNAME = os.environ.get("XPRA_HTTP_AUTH_USERNAME", "")
AUTH_PASSWORD = os.environ.get("XPRA_HTTP_AUTH_PASSWORD", "")


http_headers_cache: dict[str, str] = {}
http_headers_time: dict[str, float] = {}


def may_reload_headers(http_headers_dirs: Iterable[str]) -> dict[str, str]:
    mtimes: dict[str, float] = {}
    global http_headers_cache
    if http_headers_cache:
        # do we need to refresh the cache?
        for d in http_headers_dirs:
            if os.path.exists(d) and os.path.isdir(d):
                mtime = os.path.getmtime(d)
                if mtime > http_headers_time.get(d, -1):
                    mtimes[d] = mtime
        if not mtimes:
            return http_headers_cache.copy()
        log("headers directories have changed: %s", mtimes)
    headers: dict[str, str] = {}
    for d in http_headers_dirs:
        if not os.path.exists(d) or not os.path.isdir(d):
            continue
        mtime = os.path.getmtime(d)
        for f in sorted(os.listdir(d)):
            header_file = os.path.join(d, f)
            if not os.path.isfile(header_file):
                continue
            log("may_reload_headers() loading from '%s'", header_file)
            h: dict[str, str] = {}
            with open(header_file, encoding="latin1") as hf:
                for line in hf:
                    sline = line.strip().rstrip("\r\n").strip()
                    if sline.startswith("#") or not sline:
                        continue
                    parts = sline.split("=", 1)
                    if len(parts) != 2 and sline.find(":") > 0:
                        parts = sline.split(":", 1)
                    if len(parts) != 2:
                        continue
                    h[parts[0].strip()] = parts[1].strip()
            log(f"may_reload_headers() {header_file}={h!r}")
            headers.update(h)
        http_headers_time[d] = mtime
    log(f"may_reload_headers() headers={headers!r}, mtime={mtimes}")
    http_headers_cache = headers
    return headers.copy()


def translate_path(path: str, web_root: str = "/usr/share/xpra/www") -> str:
    # code duplicated from superclass since we can't easily inject the web_root..
    s = path
    # abandon query parameters
    path = path.split('?', 1)[0]
    path = path.split('#', 1)[0]
    # Don't forget explicit trailing slash when normalizing. Issue17324
    trailing_slash = path.rstrip().endswith('/')
    path = posixpath.normpath(unquote(path))
    words = path.split('/')
    words = list(filter(None, words))
    path = web_root
    for word in words:
        word = os.path.splitdrive(word)[1]
        word = os.path.split(word)[1]
        if word in (os.curdir, os.pardir):
            continue
        path = os.path.join(path, word)
    if trailing_slash:
        path += '/'
    # hack for locating the default desktop background at runtime:
    if not os.path.exists(path) and s.endswith("/background.png"):
        paths = get_desktop_background_paths()
        for p in paths:
            matches = glob.glob(p)
            if matches:
                path = matches[0]
                break
        if not os.path.exists(path):
            # better send something than a 404,
            # use a transparent 1x1 image:
            path = os.path.join(web_root, "icons", "empty.png")
    log("translate_path(%s)=%s", s, path)
    return path


def load_path(accept_encoding: list[str], path: str) -> tuple[int, dict[str, Any], bytes]:
    ext = os.path.splitext(path)[1]
    extra_headers: dict[str, Any] = {}
    with open(path, "rb") as f:
        # Always read in binary mode. Opening files in text mode may cause
        # newline translations, making the actual size of the content
        # transmitted *less* than the content-length!
        fs = os.fstat(f.fileno())
        content_length = fs[6]
        content_type = EXTENSION_TO_MIMETYPE.get(ext)
        if not content_type:
            if not mimetypes.inited:
                mimetypes.init()
            ctype = mimetypes.guess_type(path, False)
            if ctype and ctype[0]:
                content_type = ctype[0]
        log("guess_type(%s)=%s", path, content_type)
        if content_type:
            extra_headers["Content-type"] = content_type
        accept = tuple(accept_encoding)
        accept = tuple(x.split(";")[0].strip() for x in accept)
        content = None
        log("accept-encoding=%s", csv(accept))
        for enc in HTTP_ACCEPT_ENCODING:
            # find a matching pre-compressed file:
            if enc not in accept:
                continue
            compressed_path = f"{path}.{enc}"  # ie: "/path/to/index.html.br"
            if not os.path.exists(compressed_path):
                continue
            if not os.path.isfile(compressed_path):
                log.warn(f"Warning: {compressed_path!r} is not a file!")
                continue
            if not os.access(compressed_path, os.R_OK):
                log.warn(f"Warning: {compressed_path!r} is not readable")
                continue
            st = os.stat(compressed_path)
            if st.st_size == 0:
                log.warn(f"Warning: {compressed_path!r} is empty")
                continue
            log("sending pre-compressed file '%s'", compressed_path)
            # read pre-gzipped file:
            with open(compressed_path, "rb") as cf:
                content = cf.read()
            assert content, f"no data in {compressed_path!r}"
            extra_headers["Content-Encoding"] = enc
            break
        if not content:
            content = f.read()
            if len(content) != content_length:
                raise RuntimeError(f"expected {path!r} to contain {content_length} bytes but read {len(content)} bytes")
            if all((
                    content_length > 128,
                    "gzip" in accept,
                    "gzip" in HTTP_ACCEPT_ENCODING,
                    ext not in (".png",),
            )):
                # gzip it on the fly:
                import zlib  # pylint: disable=import-outside-toplevel
                gzip_compress = zlib.compressobj(9, zlib.DEFLATED, zlib.MAX_WBITS | 16)
                compressed_content = gzip_compress.compress(content) + gzip_compress.flush()
                if len(compressed_content) < content_length:
                    log("gzip compressed '%s': %i down to %i bytes", path, content_length, len(compressed_content))
                    extra_headers["Content-Encoding"] = "gzip"
                    content = compressed_content
        extra_headers |= {
            "Content-Length": len(content),
            "Last-Modified": fs.st_mtime,
        }
        return 200, extra_headers, content


# noinspection PyPep8Naming
class HTTPRequestHandler(BaseHTTPRequestHandler):
    """
    Xpra's builtin HTTP server.
    * locates the desktop background file at "/background.png" dynamically if missing,
    * supports the magic ?echo-headers query,
    * loads http headers from a directory (and caches the data),
    * sets cache headers on responses,
    * supports delegation to external script classes,
    * supports pre-compressed brotli and gzip, can gzip on-the-fly,
    (subclassed in WebSocketRequestHandler to add WebSocket support)
    """

    wbufsize = 0  # we flush explicitly when needed
    server_version = "Xpra-HTTP-Server"
    sys_version = "Python/" + ".".join(str(vnum) for vnum in sys.version_info[:FULL_INFO+1])

    def __init__(self, sock, addr,
                 web_root: str = "/usr/share/xpra/www/",
                 http_headers_dirs: Iterable[str] = ("/etc/xpra/http-headers", ),
                 script_paths: dict[str, Callable[[str], HttpResponse]] = None,
                 username: str = AUTH_USERNAME,
                 password: str = AUTH_PASSWORD):
        self.web_root = web_root
        self.http_headers_dirs = http_headers_dirs
        self.script_paths = script_paths or {}
        self.username = username
        self.password = password
        server = AdHocStruct()
        server.logger = log
        self.directory_listing = DIRECTORY_LISTING
        self.extra_headers: dict[str, Any] = {}
        super().__init__(sock, addr, server)

    def log_error(self, fmt, *args) -> None:  # pylint: disable=arguments-differ
        # don't log 404s at error level:
        if len(args) == 1 and isinstance(args[0], TimeoutError):
            log(fmt, *args)
        elif len(args) == 2 and args[0] == 404:
            log(fmt, *args)
        else:
            log.error(fmt, *args)

    def log_message(self, fmt, *args) -> None:  # pylint: disable=arguments-differ
        if args and len(args) == 3 and fmt == '"%s" %s %s' and args[1] == "400":
            fmt = '"%r" %s %s'
            largs = list(args)
            largs[0] = repr_ellipsized(args[0])
            args = tuple(largs)
        log(fmt, *args)

    def end_headers(self) -> None:
        # magic for querying request header values:
        path = getattr(self, "path", "")
        if path.endswith("?echo-headers"):
            # ie: "en-GB,en-US;q=0.8,en;q=0.6"
            accept = self.headers.get("Accept-Language")
            if accept:
                self.extra_headers["Echo-Accept-Language"] = std(accept, extras="-,./:;=")
        headers = self.get_headers()
        if self.extra_headers:
            headers.update(self.extra_headers)
        if headers:
            for k, v in headers.items():
                self.send_header(k, v)
        super().end_headers()

    def get_headers(self) -> dict[str, str]:
        return may_reload_headers(self.http_headers_dirs)

    def do_POST(self) -> None:
        with log.trap_error("Error processing POST request"):
            length = int(self.headers.get('content-length'))
            data = self.rfile.read(length)
            log("POST data=%s (%i bytes)", data, length)
            self.handle_request()

    def do_GET(self) -> None:
        self.handle_request()

    def handle_authentication(self) -> bool:
        if not self.password:
            return True
        authlog = Logger("auth", "http")

        def auth_err(msg):
            self.do_AUTHHEAD()
            self.wfile.write(msg.encode("latin1"))
            authlog.warn(f"http authentication failed: {msg}")
            try:
                peername = self.request.getpeername()
                authlog.warn(" from %s", pretty_socket(peername))
            except (AttributeError, OSError):
                pass
            return False

        auth = self.headers.get("Authorization")
        authlog("handle_request() auth header=%s", auth)
        if not auth:
            return auth_err("missing authentication header")
        # ie: auth = 'Basic dGVzdDp0ZXN0'
        if not auth.startswith("Basic "):
            return auth_err("invalid authentication header")
        b64str = auth.split("Basic ", 1)[1]
        import base64
        try:
            s = base64.b64decode(b64str).decode("utf8")
        except ValueError:
            s = ""
        if s.find(":") < 0:
            return auth_err("invalid authentication format")
        username, password = s.split(":", 1)
        if (self.username and username != self.username) or password != self.password:
            authlog("http authentication: expected %s:%s but received %s:%s",
                    self.username or "", self.password, username, password)
            return auth_err("invalid credentials")
        authlog("http authentication passed")
        return True

    def handle_request(self) -> None:
        if not self.handle_authentication():
            return
        content = self.send_head()
        if content:
            try:
                self.wfile.write(content)
            except (BrokenPipeError, ConnectionResetError, socket.error) as e:
                # ssl.SSLEOFError is a socket.error
                log("handle_request() %s", e)
            except TypeError:
                self.close_connection = True
                log("handle_request()", exc_info=True)
                log.error("Error handling http request")
                log.error(" for '%s'", self.path)
                log.error(" content type is %s", type(content))
            except Exception:
                self.close_connection = True
                log.error("Error handling http request")
                log.error(" for '%s'", self.path, exc_info=True)

    def do_HEAD(self) -> None:
        self.send_head()

    def do_AUTHHEAD(self) -> None:
        self.send_response(401)
        if self.password:
            self.send_header("WWW-Authenticate", f"Basic realm=\"{AUTH_REALM}\"")
        self.send_header("Content-type", "text/html")
        self.end_headers()

    def send_error(self, code, message=None, explain=None) -> None:
        try:
            super().send_error(code, message, explain)
        except OSError:
            log(f"failed to send {code} error - maybe some of the headers were already sent?", exc_info=True)

    # code taken from MIT licensed code in GzipSimpleHTTPServer.py
    def send_head(self) -> bytes:
        path = self.path.split("?", 1)[0].split("#", 1)[0]
        # strip path after second slash:
        script_path = path
        while script_path.rfind("/", 1) > 0:
            script_path = script_path[:script_path.rfind("/", 1)]
        script = self.script_paths.get(script_path)
        log("send_head() script(%s)=%s", script_path, script)
        if script:
            log("request for %s handled using %s", path, script)
            try:
                code, headers, body = script(path)
            except Exception:
                log.error(f"Error calling script {script}", exc_info=True)
                self.send_error(500, "Server error")
                return b""
            self.send_response(code)
            self.extra_headers.update(headers or {})
            self.end_headers()
            return body
        path = translate_path(self.path, self.web_root)
        if not path or not os.path.exists(path):
            self.send_error(404, "Path not found")
            return b""
        if os.path.isdir(path):
            if not path.endswith('/'):
                # redirect browser - doing basically what apache does
                self.send_response(301)
                self.send_header("Location", path + "/")
                self.end_headers()
                return b""
            for index in "index.html", "index.htm":
                index = os.path.join(path, index)
                if os.path.exists(index):
                    path = index
                    break
            else:
                if not self.directory_listing:
                    self.send_error(403, "Directory listing forbidden")
                    return b""
                code, headers, body = list_directory(path)
                self.send_response(code)
                self.extra_headers.update(headers)
                self.end_headers()
                return body
        try:
            accept_encoding = self.headers.get("accept-encoding", "").split(",")
            code, extra_headers, content = load_path(accept_encoding, path)
            lm = extra_headers.get("Last-Modified")
            if lm:
                extra_headers["Last-Modified"] = self.date_time_string(lm)
            self.send_response(code)
            self.extra_headers.update(extra_headers)
            self.end_headers()
            return content
        except OSError as e:
            self.close_connection = True
            log("send_head()", exc_info=True)
            if not isinstance(e, ConnectionResetError):
                log.error("Error sending '%s':", path)
                emsg = str(e)
                if emsg.endswith(f": '{path}'"):
                    log.error(" %s", emsg.rsplit(":", 1)[0])
                else:
                    log.estr(e)
                self.send_error(404, "File not found")
            return b""



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest
from time import monotonic
from collections import deque

try:
    from xpra.server.window import video_subregion
    from xpra import rectangle
except ImportError:
    video_subregion = None
    rectangle = None


class TestVideoSubregion(unittest.TestCase):

    def test_eq(self):
        log = video_subregion.sslog

        def refresh_cb(window, regions):
            log("refresh_cb(%s, %s)", window, regions)
        r = video_subregion.VideoSubregion(refresh_cb, 150, True)
        assert repr(r)
        r.set_detection(True)
        r.set_region(0, 0, 10, 10)
        r.set_detection(False)
        r.set_region(0, 0, 0, 0)
        r.set_detection(True)

        ww = 1024
        wh = 768
        def assertiswin():
            if not r.rectangle:
                raise Exception("region not found, should have matched the whole window")
            if r.rectangle.get_geometry()!=(0, 0, ww, wh):
                raise Exception("rectangle %s does not match whole window %ix%i" % (r.rectangle, ww, wh))

        log("* checking that we need some events")
        last_damage_events = []
        for x in range(video_subregion.MIN_EVENTS):
            last_damage_events.append((0, 0, 0, 1, 1))
        r.identify_video_subregion(ww, wh, video_subregion.MIN_EVENTS, last_damage_events)
        assert r.rectangle is None

        vr = (monotonic(), 100, 100, 320, 240)
        log("* easiest case: all updates in one region")
        last_damage_events = []
        for _ in range(50):
            last_damage_events.append(vr)
        r.identify_video_subregion(ww, wh, 50, last_damage_events)
        assert r.rectangle
        assert r.rectangle==rectangle.rectangle(*vr[1:])

        log("* checking that empty damage events does not cause errors")
        r.reset()
        r.identify_video_subregion(ww, wh, 0, [])
        assert r.rectangle is None

        log("* checking that full window can be a region")
        vr = (monotonic(), 0, 0, ww, wh)
        last_damage_events = []
        for _ in range(50):
            last_damage_events.append(vr)
        r.identify_video_subregion(ww, wh, 50, last_damage_events)
        assert r.rectangle is not None

        log("* checking that regions covering the whole window give the same result")
        last_damage_events = deque(maxlen=150)
        for x in range(4):
            for y in range(4):
                vr = (monotonic(), ww*x/4, wh*y/4, ww/4, wh/4)
                for _ in range(3):
                    last_damage_events.append(vr)
        r.identify_video_subregion(ww, wh, 150, last_damage_events)
        assertiswin()

        vr = (monotonic(), ww/4, wh/4, ww/2, wh/2)
        log("* mixed with region using 1/4 of window and 1/3 of updates: %s", vr)
        for _ in range(24):
            last_damage_events.append(vr)
        r.identify_video_subregion(ww, wh, 200, last_damage_events)
        assertiswin()

        log("* info=%s", r.get_info())

        log("* checking that two video regions quite far apart do not get merged")
        last_damage_events = deque(maxlen=150)
        r.reset()
        v1 = (monotonic(), 100, 100, 320, 240)
        v2 = (monotonic(), 500, 500, 320, 240)
        for _ in range(50):
            last_damage_events.append(v1)
            last_damage_events.append(v2)
        r.identify_video_subregion(ww, wh, 100, last_damage_events)
        assert r.rectangle is None

        log("* checking that two video regions close to each other can be merged")
        for N1, N2 in ((50, 50), (60, 40), (50, 30)):
            last_damage_events = deque(maxlen=150)
            r.reset()
            v1 = (monotonic(), 100, 100, 320, 240)
            for _ in range(N1):
                last_damage_events.append(v1)
            v2 = (monotonic(), 460, 120, 320, 240)
            for _ in range(N2):
                last_damage_events.append(v2)
            r.identify_video_subregion(ww, wh, 100, last_damage_events)
            m = rectangle.merge_all([rectangle.rectangle(*v1[1:]), rectangle.rectangle(*v2[1:])])
            assert r.rectangle and r.rectangle==m, "expected %s but got %s for N1=%i, N2=%i" % (m, r.rectangle, N1, N2)
        r.set_enabled(False)
        r.remove_refresh_region(rectangle.rectangle(0, 0, 10, 10))
        r.cleanup()

    def test_cases(self):
        from xpra.server.window.video_subregion import scoreinout   #, sslog
        #sslog.enable_debug()
        r = rectangle.rectangle(35, 435, 194, 132)
        score = scoreinout(1200, 1024, r, 1466834, 21874694)
        assert score<100
        r = rectangle.rectangle(100, 600, 320, 240)
        score = scoreinout(1200, 1024, r, 320*240*10, 320*240*25)
        assert score<100


def main():
    if video_subregion and rectangle:
        unittest.main()
    else:
        print("video_subregion_test skipped")

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest
import random
from time import monotonic

try:
    from xpra.server import cystats
except ImportError:
    cystats = None


class TestCystats(unittest.TestCase):

    def test_calculate_timesize_weighted_average(self):
        #event_time, size, elapsed_time
        now = monotonic()
        sample_size = 1000
        data = []
        ts = now - sample_size
        for _ in range(sample_size):
            s = random.randint(1000, 10000)
            v = random.random()
            data.append((ts, s, v))
            ts += 1
        a, ra = cystats.calculate_timesize_weighted_average(data)
        assert a>0 and ra>0
        #the calculations use the ratio of the size divided by the elapsed time,
        #so check that a predictable ratio gives the expected value:
        for x in (5, 1000):
            v = [(now, i*x, x) for i in range(1, 1000)]
            a, ra = cystats.calculate_size_weighted_average(v)
            #but we need to round to an int to compare
            self.assertEqual(x, round(a), "average should be %i, got %i" % (x, a))
            self.assertEqual(x, round(ra), "recent average should be %i, got %i" % (x, ra))
        def t(v, ea, era):
            a, ra = cystats.calculate_size_weighted_average(v)
            self.assertEqual(round(a), round(ea), "average should be %s, got %s" % (round(ea), round(a)))
            self.assertEqual(round(ra), round(era), "recent average should be %s, got %s" % (round(era), round(ra)))
        #an old record won't make any difference
        #compared with one that was taken just now:
        for v in (1, 10, 1000):
            #1 day ago:
            if now>60*60*24:
                t([(now-60*60*24, 1000, 1000), (now, 1000, v)], v, v)
                t([(now-60*60*24, 2*1000, 1000), (now, 1000, v)], v, v)
            #1 hour ago:
            if now>60*60:
                t([(now-60*60, 1000, 10), (now, 1000, v)], v, v)
        #but 100s ago starts to make a difference:
        t([(now-100, 1000, 1), (now, 1000, 100)], 99, 100)
        #with only 10s:
        t([(now-10, 1000, 1), (now, 1000, 100)], 92, 100)
        #1 second:
        t([(now-1, 1000, 1), (now, 1000, 100)], 67, 92)
        #if using the same time, then size matters more:
        v = [(now, 100*1000, 1000), (now, 50*1000, 1000)]
        a, ra = cystats.calculate_size_weighted_average(v)
        #recent is the same as "normal" average:
        self.assertEqual(round(a), round(ra))
        self.assertGreater(a, 75)
        #real data:
        T = monotonic()
        v =[(T-21.557095, 157684, 9110), (T-22.23345, 3744, 1279), (T-22.376621, 3744, 706),
            (T-22.515456, 3744, 1302), (T-23.013887, 78, 1342), (T-43.707768, 78, 920),
            (T-44.043399, 78, 1558), (T-44.046686, 78, 1119), (T-44.048169, 78, 1007),
            (T-44.049807, 1716, 626), (T-44.053967, 78, 2841), (T-44.23714, 78, 1393),
            (T-44.238555, 78, 2903), (T-44.242623, 78, 1167), (T-44.244426, 1716, 1032),
            (T-44.245675, 78, 720), (T-44.392009, 78, 784), (T-44.392771, 78, 737),
            (T-44.396293, 78, 911), (T-44.397466, 1716, 772), (T-44.398027, 78, 1234),
            (T-44.538323, 78, 1200), (T-44.539683, 78, 586), (T-44.542575, 78, 1203),
            (T-44.544646, 1716, 1129), (T-44.546205, 78, 979), (T-44.701881, 78, 901),
            (T-44.703987, 78, 448), (T-44.708965, 78, 474), (T-44.711481, 1716, 1444),
            (T-44.713157, 78, 1033), (T-44.848487, 78, 860), (T-44.850604, 78, 1172),
            (T-44.857039, 78, 1367), (T-44.858723, 1716, 1078), (T-44.859743, 78, 1876),
            (T-44.993883, 78, 824), (T-44.99714, 78, 796), (T-45.001942, 78, 714),
            (T-45.002884, 1716, 744), (T-45.004841, 78, 652), (T-45.772856, 78, 652)]
        raw_v = [x[2] for x in v]
        min_v = min(raw_v)
        max_v = max(raw_v)
        a, ra = cystats.calculate_size_weighted_average(v)
        self.assertLess(a, max_v)
        self.assertLess(ra, max_v)
        self.assertGreater(a, min_v)
        self.assertGreater(ra, min_v)

    def test_calculate_time_weighted_average(self):
        now = monotonic()
        sample_size = 100
        data = []
        t = now - sample_size
        for _ in range(sample_size):
            v = random.random()
            data.append((t, v))
            t += 1
        a, ra = cystats.calculate_time_weighted_average(data)
        assert 0<a<1 and 0<ra<1

    def test_logp(self):
        for _ in range(1000):
            x = random.random()
            v = cystats.logp(x)
            assert 0<=v<=1
        for x in (0, 1):
            v = cystats.logp(x)
            assert 0<=v<=1


def main():
    if cystats:
        unittest.main()

if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import gi

gi.require_version('Gtk', '3.0')  # @UndefinedVariable
from gi.repository import Gtk  #pylint: disable=wrong-import-position @UnresolvedImport

from xpra.x11.gtk.display_source import init_gdk_display_source

init_gdk_display_source()


def change_callback(self, window, entry):
    print("content_type=%s" % entry.get_text())
    if window.get_window():
        from xpra.x11.gtk.prop import prop_set
        prop_set(window.get_window(), "_XPRA_CONTENT_TYPE", "latin1", entry.get_text().decode())


def main():
    window = Gtk.Window(type=Gtk.WindowType.TOPLEVEL)
    window.set_size_request(400, 100)
    window.connect("delete_event", Gtk.main_quit)
    entry = Gtk.Entry()
    entry.set_max_length(50)
    entry.connect("changed", change_callback, window, entry)
    content_type = "text"
    import sys
    if len(sys.argv) > 1:
        content_type = sys.argv[1]
    entry.set_text(content_type)
    entry.show()
    window.add(entry)
    window.show_all()
    Gtk.main()
    return 0


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest

from xpra.util.objects import typedict, AdHocStruct
from xpra.client.base.serverinfo import ServerInfoMixin
from unit.client.mixins.clientmixintest_util import ClientMixinTest


class ServerInfoClientTest(ClientMixinTest):

	def test_audio(self):
		waq = []
		def _ServerInfoMixin():
			x = ServerInfoMixin()
			def warn_and_quit(*args):
				waq.append(args)
			x.warn_and_quit = warn_and_quit
			return x
		opts = AdHocStruct()
		caps = typedict({
			"machine_id" : "123",
			"uuid"	: "some-uuid",
			"build.version"	: "3.0",
			"build.revision" : "23000",
			"hostname"	: "localhost",
			"display" : ":99",
			"platform" : "linux2",
			"platform.release" : "dunno",
			"platform.platform" : "platformX",
			"platform.linux_distribution" : ('Linux Fedora', 20, 'Heisenbug'),
			})
		x = self._test_mixin_class(_ServerInfoMixin, opts, caps)
		del caps["build.version"]
		assert not x.parse_server_capabilities(caps), "should have failed when version is missing"
		version = "0.1"
		caps["build.version"] = version
		assert not x.parse_server_capabilities(caps), "should have failed with version %s" % version


def main():
	unittest.main()


if __name__ == '__main__':
	main()
